#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
General Kubernetes API Client Example
Can be used to perform various Kubernetes API operations
"""
import logging
import os

import yaml
from dotenv import load_dotenv
from kubernetes import client, config
from kubernetes.client import V1DeleteOptions
from kubernetes.client.rest import ApiException


class KubernetesApiClient:
    """Kubernetes API Client Wrapper Class"""

    def __init__(self, kubeconfig_path=None, context=None, in_cluster=False):
        """
        Initialize Kubernetes API Client

        Args:
            kubeconfig_path: kubeconfig file path, defaults to None which uses ~/.kube/config
            context: kubeconfig context name to use
            in_cluster: whether running inside a cluster, if True use service account configuration
        """
        try:
            # Use absolute path relative to the script file for KUBECONFIG_PATH
            load_dotenv()
            script_dir = os.path.dirname(os.path.abspath(__file__))
            script_path = os.path.join(script_dir, "kubeconfig")
            kubeconfig_path = kubeconfig_path or os.getenv("KUBECONFIG_PATH") or script_path
            if in_cluster:
                config.load_incluster_config()
            else:
                config.load_kube_config(
                    config_file=kubeconfig_path,
                    context=context
                )

            # Initialize various API clients
            self.core_v1 = client.CoreV1Api()
            self.apps_v1 = client.AppsV1Api()
            self.batch_v1 = client.BatchV1Api()
            self.networking_v1 = client.NetworkingV1Api()
            self.rbac_v1 = client.RbacAuthorizationV1Api()
            self.custom_objects = client.CustomObjectsApi()

            logging.info("Kubernetes API client initialized successfully")
        except Exception as e:
            logging.info(f"Failed to initialize Kubernetes API client: {e}")
            raise

    # ===================== Pod Operations =====================

    def get_pod(self, name, namespace="default"):
        """
        Get a specific Pod in the given namespace
        Equivalent to: GET /api/v1/namespaces/{namespace}/pods/{name}
        """
        try:
            return self.core_v1.read_namespaced_pod(
                name=name,
                namespace=namespace
            )
        except ApiException as e:
            logging.warning(f"Failed to get Pod {namespace}/{name}: {e}")
            return None

    def list_pods(self, namespace="default", label_selector=None, field_selector=None):
        """
        List all Pods in the given namespace
        Equivalent to: GET /api/v1/namespaces/{namespace}/pods
        """
        try:
            return self.core_v1.list_namespaced_pod(
                namespace=namespace,
                label_selector=label_selector,
                field_selector=field_selector
            )
        except ApiException as e:
            logging.warning(f"Failed to list Pods in namespace {namespace}: {e}")
            return None

    def list_pods_all_namespaces(self, label_selector=None, field_selector=None):
        """
        List Pods across all namespaces
        Equivalent to: GET /api/v1/pods
        """
        try:
            return self.core_v1.list_pod_for_all_namespaces(
                label_selector=label_selector,
                field_selector=field_selector
            )
        except ApiException as e:
            logging.warning(f"Failed to list Pods across all namespaces: {e}")
            return None

    def create_pod(self, pod_manifest, namespace="default"):
        """
        Create a Pod
        Equivalent to: POST /api/v1/namespaces/{namespace}/pods

        Args:
            pod_manifest: Pod resource definition, can be dict or V1Pod object
            namespace: Namespace where the Pod will be created

        Returns:
            V1Pod: The created Pod object on success
            None: On failure
        """
        try:
            # If input is a dictionary, use it directly
            if isinstance(pod_manifest, dict):
                # Using dictionary definition
                return self.core_v1.create_namespaced_pod(
                    namespace=namespace,
                    body=pod_manifest
                )
            else:
                # Using V1Pod object directly
                return self.core_v1.create_namespaced_pod(
                    namespace=namespace,
                    body=pod_manifest
                )
        except ApiException as e:
            logging.warning(f"Failed to create Pod: {e}")
            return None

    def create_pod_from_yaml(self, yaml_file=None, namespace=None, pod_name=None):
        """
        Create a Pod from YAML file

        Args:
            yaml_file: YAML file path
            namespace: Namespace where the Pod will be created
            pod_name: Override the Pod name in the YAML

        Returns:
            V1Pod: The created Pod object on success
            None: On failure
        """
        try:
            load_dotenv()
            script_dir = os.path.dirname(os.path.abspath(__file__))
            pod_path = os.path.join(script_dir, "pod.yaml")
            yaml_file = yaml_file or os.getenv("POD_YAML_PATH") or pod_path
            namespace = namespace or os.getenv("POD_NAMESPACE") or "default"
            with open(yaml_file, 'r') as f:
                pod_manifest = yaml.safe_load(f)
            # Update Pod name if provided
            if pod_name:
                if 'metadata' in pod_manifest:
                    pod_manifest['metadata']['name'] = pod_name
                    if 'labels' in pod_manifest['metadata']:
                        pod_manifest['metadata']['labels']['name'] = pod_name
                if 'spec' in pod_manifest and 'containers' in pod_manifest['spec'] and pod_manifest['spec'][
                    'containers']:
                    pod_manifest['spec']['containers'][0]['name'] = pod_name

            return self.create_pod(pod_manifest, namespace)
        except Exception as e:
            logging.info(f"Failed to create Pod from YAML file: {e}")
            return None

    def delete_pod(self, name, namespace="default", grace_period_seconds=30):
        """
        Delete a Pod
        Equivalent to: DELETE /api/v1/namespaces/{namespace}/pods/{name}

        Args:
            name: Pod name
            namespace: Namespace where the Pod is located
            grace_period_seconds: Grace period in seconds

        Returns:
            V1Status: Status object on successful deletion
            None: On failure
        """
        try:
            return self.core_v1.delete_namespaced_pod(
                name=name,
                namespace=namespace,
                body=V1DeleteOptions(
                    grace_period_seconds=grace_period_seconds,
                    propagation_policy="Background"
                )
            )
        except ApiException as e:
            logging.warning(f"Failed to delete Pod {namespace}/{name}: {e}")
            return None

    def update_pod(self, name, pod_manifest, namespace="default"):
        """
        Update a Pod
        Equivalent to: PUT /api/v1/namespaces/{namespace}/pods/{name}

        Args:
            name: Pod name
            pod_manifest: Pod resource definition, can be dict or V1Pod object
            namespace: Namespace where the Pod is located

        Returns:
            V1Pod: The updated Pod object on success
            None: On failure
        """
        try:
            # If input is a dictionary, ensure name and namespace fields
            if isinstance(pod_manifest, dict):
                if 'metadata' not in pod_manifest:
                    pod_manifest['metadata'] = {}
                pod_manifest['metadata']['name'] = name
                pod_manifest['metadata']['namespace'] = namespace

                return self.core_v1.replace_namespaced_pod(
                    name=name,
                    namespace=namespace,
                    body=pod_manifest
                )
            else:
                # Ensure V1Pod object has correct name and namespace
                pod_manifest.metadata.name = name
                pod_manifest.metadata.namespace = namespace

                return self.core_v1.replace_namespaced_pod(
                    name=name,
                    namespace=namespace,
                    body=pod_manifest
                )
        except ApiException as e:
            logging.warning(f"Failed to update Pod {namespace}/{name}: {e}")
            return None

    def patch_pod(self, name, patch_data, namespace="default"):
        """
        Partially update a Pod
        Equivalent to: PATCH /api/v1/namespaces/{namespace}/pods/{name}

        Args:
            name: Pod name
            patch_data: Data to update, in dictionary format
            namespace: Namespace where the Pod is located

        Returns:
            V1Pod: The updated Pod object on success
            None: On failure
        """
        try:
            return self.core_v1.patch_namespaced_pod(
                name=name,
                namespace=namespace,
                body=patch_data
            )
        except ApiException as e:
            logging.warning(f"Failed to patch Pod {namespace}/{name}: {e}")
            return None

    def get_pod_info(self, name, namespace="default"):
        """
        Get basic information about a Pod

        Args:
            name: Pod name
            namespace: Namespace where the Pod is located

        Returns:
            dict: Dictionary containing basic Pod information including status, IP, start time, etc.
            None: On failure
        """
        try:
            pod = self.get_pod(name, namespace)
            if not pod:
                return None

            # Format start time for readability
            start_time = None
            if pod.status.start_time:
                # Convert time to readable format (ISO format: YYYY-MM-DD HH:MM:SS)
                start_time_obj = pod.status.start_time.replace(tzinfo=None)
                start_time = start_time_obj.strftime('%Y-%m-%d %H:%M:%S')

            pod_info = {
                "pod_name": pod.metadata.name,
                "namespace": pod.metadata.namespace,
                "status": pod.status.phase,  # Pending, Running, Succeeded, Failed, Unknown
                "pod_ip": pod.status.pod_ip,
                "host_ip": pod.status.host_ip,
                "start_time": start_time,
                "node_name": pod.spec.node_name if hasattr(pod.spec, "node_name") else None
            }

            return pod_info
        except Exception as e:
            logging.warning(f"Failed to get Pod information for {namespace}/{name}: {e}")
            return None

    # ===================== Deployment Operations =====================

    def get_deployment(self, name, namespace="default"):
        """
        Get a specific Deployment
        Equivalent to: GET /apis/apps/v1/namespaces/{namespace}/deployments/{name}
        """
        try:
            return self.apps_v1.read_namespaced_deployment(
                name=name,
                namespace=namespace
            )
        except ApiException as e:
            logging.warning(f"Failed to get Deployment {namespace}/{name}: {e}")
            return None

    def list_deployments(self, namespace="default", label_selector=None):
        """
        List all Deployments in the given namespace
        Equivalent to: GET /apis/apps/v1/namespaces/{namespace}/deployments
        """
        try:
            return self.apps_v1.list_namespaced_deployment(
                namespace=namespace,
                label_selector=label_selector
            )
        except ApiException as e:
            logging.warning(f"Failed to list Deployments in namespace {namespace}: {e}")
            return None

    # ===================== Service Operations =====================

    def get_service(self, name, namespace="default"):
        """
        Get a specific Service
        Equivalent to: GET /api/v1/namespaces/{namespace}/services/{name}
        """
        try:
            return self.core_v1.read_namespaced_service(
                name=name,
                namespace=namespace
            )
        except ApiException as e:
            logging.warning(f"Failed to get Service {namespace}/{name}: {e}")
            return None

    def list_services(self, namespace="default", label_selector=None):
        """
        List all Services in the given namespace
        Equivalent to: GET /api/v1/namespaces/{namespace}/services
        """
        try:
            return self.core_v1.list_namespaced_service(
                namespace=namespace,
                label_selector=label_selector
            )
        except ApiException as e:
            logging.warning(f"Failed to list Services in namespace {namespace}: {e}")
            return None

    def create_service(self, service_manifest, namespace="default"):
        """
        Create a Service
        Equivalent to: POST /api/v1/namespaces/{namespace}/services

        Args:
            service_manifest: Service resource definition, can be dict or V1Service object
            namespace: Namespace where the Service will be created

        Returns:
            V1Service: The created Service object on success
            None: On failure
        """
        try:
            # If input is a dictionary, use it directly
            if isinstance(service_manifest, dict):
                # Using dictionary definition
                return self.core_v1.create_namespaced_service(
                    namespace=namespace,
                    body=service_manifest
                )
            else:
                # Using V1Service object directly
                return self.core_v1.create_namespaced_service(
                    namespace=namespace,
                    body=service_manifest
                )
        except ApiException as e:
            logging.warning(f"Failed to create Service: {e}")
            return None

    def create_service_from_yaml(self, yaml_file=None, namespace=None, service_name=None, selector_name=None):
        """
        Create a Service from YAML file

        Args:
            yaml_file: YAML file path
            namespace: Namespace where the Service will be created
            service_name: Override the Service name in the YAML
            selector_name: Override the selector name for pod targeting

        Returns:
            V1Service: The created Service object on success
            None: On failure
        """
        try:
            load_dotenv()
            script_dir = os.path.dirname(os.path.abspath(__file__))
            service_path = os.path.join(script_dir, "service.yaml")
            yaml_file = yaml_file or os.getenv("SERVICE_YAML_PATH") or service_path
            namespace = namespace or os.getenv("SERVICE_NAMESPACE") or "default"
            with open(yaml_file, 'r') as f:
                service_manifest = yaml.safe_load(f)

            # Update Service name if provided
            if service_name:
                if 'metadata' in service_manifest:
                    service_manifest['metadata']['name'] = service_name
                    # Update app label if present
                    if 'labels' in service_manifest['metadata']:
                        service_manifest['metadata']['labels']['app'] = service_name

            # Update selector if provided
            if selector_name:
                if 'spec' in service_manifest and 'selector' in service_manifest['spec']:
                    service_manifest['spec']['selector']['name'] = selector_name

            return self.create_service(service_manifest, namespace)
        except Exception as e:
            logging.info(f"Failed to create Service from YAML file: {e}")
            return None

    def update_service(self, name, service_manifest, namespace="default"):
        """
        Update a Service
        Equivalent to: PUT /api/v1/namespaces/{namespace}/services/{name}

        Args:
            name: Service name
            service_manifest: Service resource definition, can be dict or V1Service object
            namespace: Namespace where the Service is located

        Returns:
            V1Service: The updated Service object on success
            None: On failure
        """
        try:
            # If input is a dictionary, ensure name and namespace fields
            if isinstance(service_manifest, dict):
                if 'metadata' not in service_manifest:
                    service_manifest['metadata'] = {}
                service_manifest['metadata']['name'] = name
                service_manifest['metadata']['namespace'] = namespace

                return self.core_v1.replace_namespaced_service(
                    name=name,
                    namespace=namespace,
                    body=service_manifest
                )
            else:
                # Ensure V1Service object has correct name and namespace
                service_manifest.metadata.name = name
                service_manifest.metadata.namespace = namespace

                return self.core_v1.replace_namespaced_service(
                    name=name,
                    namespace=namespace,
                    body=service_manifest
                )
        except ApiException as e:
            logging.warning(f"Failed to update Service {namespace}/{name}: {e}")
            return None

    def patch_service(self, name, patch_data, namespace="default"):
        """
        Partially update a Service
        Equivalent to: PATCH /api/v1/namespaces/{namespace}/services/{name}

        Args:
            name: Service name
            patch_data: Data to update, in dictionary format
            namespace: Namespace where the Service is located

        Returns:
            V1Service: The updated Service object on success
            None: On failure
        """
        try:
            return self.core_v1.patch_namespaced_service(
                name=name,
                namespace=namespace,
                body=patch_data
            )
        except ApiException as e:
            logging.warning(f"Failed to patch Service {namespace}/{name}: {e}")
            return None

    def delete_service(self, name, namespace="default"):
        """
        Delete a Service
        Equivalent to: DELETE /api/v1/namespaces/{namespace}/services/{name}

        Args:
            name: Service name
            namespace: Namespace where the Service is located

        Returns:
            V1Status: Status object on successful deletion
            None: On failure
        """
        try:
            return self.core_v1.delete_namespaced_service(
                name=name,
                namespace=namespace
            )
        except ApiException as e:
            logging.warning(f"Failed to delete Service {namespace}/{name}: {e}")
            return None

    def get_service_info(self, name, namespace="default"):
        """
        Get basic information about a Service

        Args:
            name: Service name
            namespace: Namespace where the Service is located

        Returns:
            dict: Dictionary containing basic Service information including type, IP, ports, etc.
            None: On failure
        """
        try:
            service = self.get_service(name, namespace)
            if not service:
                return None

            # Format creation time for readability
            creation_time = None
            if service.metadata.creation_timestamp:
                creation_time_obj = service.metadata.creation_timestamp.replace(tzinfo=None)
                creation_time = creation_time_obj.strftime('%Y-%m-%d %H:%M:%S')

            # Simplify port information
            ports_info = []
            if service.spec.ports:
                for port in service.spec.ports:
                    port_info = {
                        "port": port.port,
                        "target_port": port.target_port,
                        "protocol": port.protocol
                    }

                    # Add node port if present for NodePort type
                    if hasattr(port, "node_port") and port.node_port:
                        port_info["node_port"] = port.node_port

                    ports_info.append(port_info)

            service_info = {
                "service_name": service.metadata.name,
                "namespace": service.metadata.namespace,
                "type": service.spec.type,  # ClusterIP, NodePort, LoadBalancer, ExternalName
                "cluster_ip": service.spec.cluster_ip,
                "creation_time": creation_time,
                "ports": ports_info,
                "selector": service.spec.selector,
                "host": ''  # Default to use cluster_ip as host
            }

            # Add external IP information for LoadBalancer type
            if service.spec.type == "LoadBalancer" and hasattr(service.status,
                                                               "load_balancer") and service.status.load_balancer:
                external_ips = []
                if hasattr(service.status.load_balancer, "ingress") and service.status.load_balancer.ingress:
                    for ingress in service.status.load_balancer.ingress:
                        if hasattr(ingress, "ip") and ingress.ip:
                            external_ips.append(ingress.ip)
                        elif hasattr(ingress, "hostname") and ingress.hostname:
                            external_ips.append(ingress.hostname)

                    # If there are external IPs or hostnames, use the first one as the host field
                    if external_ips:
                        service_info["host"] = external_ips[0]

                service_info["external_ips"] = external_ips

            # Add external name for ExternalName type
            if service.spec.type == "ExternalName" and hasattr(service.spec, "external_name"):
                service_info["external_name"] = service.spec.external_name
                service_info["host"] = service.spec.external_name  # For ExternalName type, use external_name as host

            return service_info
        except Exception as e:
            print(f"Failed to get Service information for {namespace}/{name}: {e}")
            return None

    # ===================== Namespace Operations =====================

    def get_namespace(self, name):
        """
        Get a specific namespace
        Equivalent to: GET /api/v1/namespaces/{name}
        """
        try:
            return self.core_v1.read_namespace(name=name)
        except ApiException as e:
            logging.warning(f"Failed to get namespace {name}: {e}")
            return None

    def list_namespaces(self, label_selector=None):
        """
        List all namespaces
        Equivalent to: GET /api/v1/namespaces
        """
        try:
            return self.core_v1.list_namespace(label_selector=label_selector)
        except ApiException as e:
            logging.warning(f"Failed to list namespaces: {e}")
            return None

    # ===================== Custom Resource (CRD) Operations =====================

    def get_custom_resource(self, group, version, plural, name, namespace=None):
        """
        Get a custom resource

        For namespaced resources:
        GET /apis/{group}/{version}/namespaces/{namespace}/{plural}/{name}

        For cluster-scoped resources:
        GET /apis/{group}/{version}/{plural}/{name}
        """
        try:
            if namespace:
                return self.custom_objects.get_namespaced_custom_object(
                    group=group,
                    version=version,
                    namespace=namespace,
                    plural=plural,
                    name=name
                )
            else:
                return self.custom_objects.get_cluster_custom_object(
                    group=group,
                    version=version,
                    plural=plural,
                    name=name
                )
        except ApiException as e:
            resource_path = f"{namespace}/{name}" if namespace else name
            logging.warning(f"Failed to get custom resource {group}/{version}/{plural}/{resource_path}: {e}")
            return None


if __name__ == "__main__":
    client = KubernetesApiClient("./kubeconfig")
    # result = client.get_pod("mcp-openapi-node-2", namespace="default")
    # print(result)
    #result = client.create_pod_from_yaml("./pod.yaml")

    # result = client.get_pod_info("mcp-openapi-node-5", namespace="default")
    # print(result)
    result = client.get_service_info("mcp-openapi-service-1", namespace="default")
    print(result)

    # result = client.create_pod_from_yaml("./pod.yaml")
    # print(result)

    # result = client.create_service_from_yaml("./service.yaml")
    # print(result)

    # result = client.delete_pod("mcp-openapi-node-1")
    # print(result)

    # result = client.delete_service("mcp-openapi-service-2")
    # print(result)




from aworld.output.base import Output, SearchOutput, SearchItem, ToolResultOutput, MessageOutput, ToolCallOutput, \
    RUN_FINISHED_SIGNAL
from aworld.output.artifact import Artifact, ArtifactType
from aworld.output.code_artifact import CodeArtifact, ShellArtifact
from aworld.output.outputs import Outputs, StreamingOutputs
from aworld.output.workspace import WorkSpace
from aworld.output.observer import WorkspaceObserver,get_observer
from aworld.output.storage.artifact_repository import ArtifactRepository, LocalArtifactRepository
from aworld.output.ui.base import AworldUI,PrinterAworldUI
__all__ = [
    "Output",
    "Artifact",
    "ArtifactType",
    "CodeArtifact",
    "ShellArtifact",
    "WorkSpace",
    "ArtifactRepository",
    "LocalArtifactRepository",
    "WorkspaceObserver",
    "get_observer",
    "SearchOutput",
    "SearchItem",
    "MessageOutput",
    "ToolCallOutput",
    "ToolResultOutput",
    "Outputs",
    "StreamingOutputs",
    "RUN_FINISHED_SIGNAL",
    "AworldUI",
    "PrinterAworldUI"
]


import asyncio
import os
import threading
import copy
import json
import openai
from importlib.metadata import version
from aworld.logs.util import logger
from aworld.trace.base import Span
from aworld.utils import import_package

_PYDANTIC_VERSION = version("pydantic")


def should_trace_prompts():
    '''Determine whether it is necessary to record the message
    '''
    return (os.getenv("SHOULD_TRACE_PROMPTS") or "true").lower() == "true"


def run_async(method):
    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        loop = None

    if loop and loop.is_running():
        thread = threading.Thread(target=lambda: asyncio.run(method))
        thread.start()
        thread.join()
    else:
        asyncio.run(method)


async def handle_openai_request(span: Span, kwargs, instance):
    if not span or not span.is_recording():
        return
    try:
        attributes = parser_request_params(kwargs, instance)
        if should_trace_prompts():
            messages = kwargs.get("messages")
            for i, msg in enumerate(messages):
                prefix = f"llm.prompts.{i}"
                attributes.update({f"{prefix}.role": msg.get("role")})
                if msg.get("content"):
                    content = copy.deepcopy(msg.get("content"))
                    content = json.dumps(content)
                    attributes.update({f"{prefix}.content": content})
                if msg.get("tool_call_id"):
                    attributes.update({
                        f"{prefix}.tool_call_id": msg.get("tool_call_id")})
                tool_calls = msg.get("tool_calls")
                if tool_calls:
                    for i, tool_call in enumerate(tool_calls):
                        tool_call = model_as_dict(tool_call)
                        function = tool_call.get("function")
                        attributes.update({
                            f"{prefix}.tool_calls.{i}.id": tool_call.get("id")})
                        attributes.update({
                            f"{prefix}.tool_calls.{i}.name": function.get("name")})
                        attributes.update({
                            f"{prefix}.tool_calls.{i}.arguments": function.get("arguments")})

        span.set_attributes(attributes)
    except ValueError as e:
        logger.warning(f"trace handle openai request error: {e}")


def parser_request_params(kwargs, instance):
    attributes = {
        "llm.system": "OpenAI",
        "llm.model": kwargs.get("model", ""),
        "llm.max_tokens": kwargs.get("max_tokens", ""),
        "llm.temperature": kwargs.get("temperature", ""),
        "llm.top_p": kwargs.get("top_p", ""),
        "llm.frequency_penalty": kwargs.get("frequency_penalty", ""),
        "llm.presence_penalty": kwargs.get("presence_penalty", ""),
        "llm.user": kwargs.get("user", ""),
        "llm.headers": kwargs.get("headers", ""),
        "llm.extra_headers": kwargs.get("extra_headers", ""),
        "llm.stream": kwargs.get("stream", "")
    }

    client = instance._client
    if isinstance(client, (openai.AsyncOpenAI, openai.OpenAI)):
        attributes.update({"llm.base_url": str(client.base_url)})

    filterd_attri = {k: v for k, v in attributes.items()
                     if (v and v is not "")}
    return filterd_attri


def is_streaming_response(response):
    return isinstance(response, openai.Stream) or isinstance(response, openai.AsyncStream)


def parse_openai_response(response, request_kwargs, instance, is_streaming):
    return {
        "llm.system": "OpenAI",
        "llm.response.model": response.get("model") or request_kwargs.get("model") or None,
        "llm.operation.name": "chat",
        "llm.server.address": _get_openai_base_url(instance),
        "llm.stream": is_streaming,
    }


def record_stream_token_usage(complete_response, request_kwargs) -> tuple[int, int]:
    '''
        return (prompt_usage, completion_usage)
    '''
    prompt_usage = 0
    completion_usage = 0

    # prompt_usage
    if request_kwargs and request_kwargs.get("messages"):
        prompt_content = ""
        model_name = complete_response.get(
            "model") or request_kwargs.get("model") or "gpt-4"
        for msg in request_kwargs.get("messages"):
            if msg.get("content"):
                prompt_content += msg.get("content")
        if model_name:
            prompt_usage = get_token_count_from_string(
                prompt_content, model_name)

    # completion_usage
    if complete_response.get("choices"):
        completion_content = ""
        model_name = complete_response.get("model") or "gpt-4"

        for choice in complete_response.get("choices"):
            if choice.get("message") and choice.get("message").get("content"):
                completion_content += choice["message"]["content"]

        if model_name:
            completion_usage = get_token_count_from_string(
                completion_content, model_name)

    return (prompt_usage, completion_usage)


def _get_openai_base_url(instance):
    if hasattr(instance, "_client"):
        client = instance._client  # pylint: disable=protected-access
        if isinstance(client, (openai.AsyncOpenAI, openai.OpenAI)):
            return str(client.base_url)

    return ""


def get_token_count_from_string(string: str, model_name: str):
    import_package("tiktoken")
    import tiktoken

    if tiktoken_encodings.get(model_name) is None:
        try:
            encoding = tiktoken.encoding_for_model(model_name)
        except KeyError as ex:
            logger.warning(
                f"Failed to get tiktoken encoding for model_name {model_name}, error: {str(ex)}")
            return None

        tiktoken_encodings[model_name] = encoding
    else:
        encoding = tiktoken_encodings.get(model_name)

    token_count = len(encoding.encode(string))
    return token_count


def record_stream_response_chunk(chunk, complete_response):
    chunk = model_as_dict(chunk)
    complete_response["model"] = chunk.get("model")
    complete_response["id"] = chunk.get("id")

    # prompt filter results
    if chunk.get("prompt_filter_results"):
        complete_response["prompt_filter_results"] = chunk.get(
            "prompt_filter_results")

    for choice in chunk.get("choices"):
        index = choice.get("index")
        if len(complete_response.get("choices")) <= index:
            complete_response["choices"].append(
                {"index": index, "message": {"content": "", "role": ""}})
        complete_choice = complete_response.get("choices")[index]
        if choice.get("finish_reason"):
            complete_choice["finish_reason"] = choice.get("finish_reason")
        if choice.get("content_filter_results"):
            complete_choice["content_filter_results"] = choice.get(
                "content_filter_results")

        delta = choice.get("delta")

        if delta and delta.get("content"):
            complete_choice["message"]["content"] += delta.get("content")

        if delta and delta.get("role"):
            complete_choice["message"]["role"] = delta.get("role")
        if delta and delta.get("tool_calls"):
            tool_calls = delta.get("tool_calls")
            if not isinstance(tool_calls, list) or len(tool_calls) == 0:
                continue

            if not complete_choice["message"].get("tool_calls"):
                complete_choice["message"]["tool_calls"] = []

            for tool_call in tool_calls:
                i = int(tool_call["index"])
                if len(complete_choice["message"]["tool_calls"]) <= i:
                    complete_choice["message"]["tool_calls"].append(
                        {"id": "", "function": {"name": "", "arguments": ""}}
                    )

                span_tool_call = complete_choice["message"]["tool_calls"][i]
                span_function = span_tool_call["function"]
                tool_call_function = tool_call.get("function")

                if tool_call.get("id"):
                    span_tool_call["id"] = tool_call.get("id")
                if tool_call_function and tool_call_function.get("name"):
                    span_function["name"] = tool_call_function.get("name")
                if tool_call_function and tool_call_function.get("arguments"):
                    span_function["arguments"] += tool_call_function.get(
                        "arguments")


def parse_response_message(choices) -> dict:
    attributes = {}
    if not should_trace_prompts():
        return attributes
    for choice in choices:
        index = choice.get("index")
        prefix = f"llm.completions.{index}"
        attributes.update(
            {f"{prefix}.finish_reason": choice.get("finish_reason")})

        message = choice.get("message")
        if not message:
            continue

        attributes.update({f"{prefix}.role": message.get("role")})

        if message.get("refusal"):
            attributes.update({f"{prefix}.refusal": message.get("refusal")})
        else:
            attributes.update({f"{prefix}.content": message.get("content")})

        function_call = message.get("function_call")
        if function_call:
            attributes.update(
                {f"{prefix}.tool_calls.0.name": function_call.get("name")})
            attributes.update(
                {f"{prefix}.tool_calls.0.arguments": function_call.get("arguments")})

        tool_calls = message.get("tool_calls")
        if tool_calls:
            for i, tool_call in enumerate(tool_calls):
                function = tool_call.get("function")
                attributes.update(
                    {f"{prefix}.tool_calls.{i}.id": tool_call.get("id")})
                attributes.update(
                    {f"{prefix}.tool_calls.{i}.name": function.get("name")})
                attributes.update(
                    {f"{prefix}.tool_calls.{i}.arguments": function.get("arguments")})
    return attributes


def model_as_dict(model):
    if isinstance(model, dict):
        return model
    if _PYDANTIC_VERSION < "2.0.0":
        return model.dict()
    if hasattr(model, "model_dump"):
        return model.model_dump()
    elif hasattr(model, "parse"):  # Raw API response
        return model_as_dict(model.parse())
    else:
        return model



import uuid
from datetime import datetime
from enum import Enum, auto
from typing import Dict, Any, Optional, ClassVar
from pydantic import Field, field_validator, model_validator

from aworld.output.base import Output


class ArtifactType(Enum):
    """Defines supported artifact types"""
    TEXT = "TEXT"
    CODE = "CODE"
    MARKDOWN = "MARKDOWN"
    HTML = "HTML"
    SVG = "SVG"
    JSON = "JSON"
    CSV = "CSV"
    TABLE = "TABLE"
    CHART = "CHART"
    DIAGRAM = "DIAGRAM"
    MCP_CALL = "MCP_CALL"
    TOOL_CALL = "TOOL_CALL"
    LLM_OUTPUT = "LLM_OUTPUT"
    WEB_PAGES = "WEB_PAGES"
    DIR = "DIR"
    CUSTOM = "CUSTOM"



class ArtifactStatus(Enum):
    """Artifact status"""
    DRAFT = auto()      # Draft status
    COMPLETE = auto()   # Completed status
    EDITED = auto()     # Edited status
    ARCHIVED = auto()   # Archived status

class Artifact(Output):
    """
    Represents a specific content generation result (artifact)
    
    Artifacts are the basic units of Artifacts technology, representing a structured content unit
    Can be code, markdown, charts, and various other formats
    """

    artifact_id: str = Field(default_factory=lambda: str(uuid.uuid4()), description="Unique identifier for the artifact")
    artifact_type: ArtifactType = Field(..., description="Type of the artifact")
    content: Any = Field(..., description="Content of the artifact")
    metadata: Dict[str, Any] = Field(default_factory=dict, description="Metadata associated with the artifact")
    created_at: str = Field(default_factory=lambda: datetime.now().isoformat(), description="Creation timestamp")
    updated_at: str = Field(default_factory=lambda: datetime.now().isoformat(), description="Last updated timestamp")
    status: ArtifactStatus = Field(default=ArtifactStatus.ARCHIVED, description="Current status of the artifact")
    current_version: str = Field(default="", description="Current version of the artifact")
    version_history: list = Field(default_factory=list, description="History of versions for the artifact")
    create_file: bool = Field(default=False, description="Flag to indicate if a file should be created")

    # Use model_validator for initialization logic
    @model_validator(mode='after')
    def setup_artifact(self):
        """Initialize the artifact after validation"""
        # Ensure artifact_id is always a valid string
        if not self.artifact_id:
            self.artifact_id = str(uuid.uuid4())
            
        # Reset status to DRAFT for new artifacts
        if not self.version_history:
            self.status = ArtifactStatus.DRAFT
            # Record initial version
            self._record_version("Initial version")
        
        return self

    def _record_version(self, description: str) -> None:
        """Record current state as a new version"""
        version = {
            "timestamp": datetime.now().isoformat(),
            "description": description,
            "content": self.content,
            "status": self.status
        }
        self.version_history.append(version)
        self.updated_at = version["timestamp"]

    def update_content(self, content: Any, description: str = "Content update") -> None:
        """
        Update artifact content and record version
        
        Args:
            content: New content
            description: Update description
        """
        self.content = content
        self.status = ArtifactStatus.EDITED
        self._record_version(description)

    def update_metadata(self, metadata: Dict[str, Any]) -> None:
        """
        Update artifact metadata
        
        Args:
            metadata: New metadata (will be merged with existing metadata)
        """
        self.metadata.update(metadata)
        self.updated_at = datetime.now().isoformat()

    def mark_complete(self) -> None:
        """Mark the artifact as complete"""
        self.status = ArtifactStatus.COMPLETE
        self._record_version("Marked as complete")

    def archive(self) -> None:
        """Archive the artifact"""
        self.status = ArtifactStatus.ARCHIVED
        self._record_version("Artifact archived")

    def get_version(self, index: int) -> Optional[Dict[str, Any]]:
        """Get version at the specified index"""
        if 0 <= index < len(self.version_history):
            return self.version_history[index]
        return None

    def revert_to_version(self, index: int) -> bool:
        """Revert to a specific version"""
        version = self.get_version(index)
        if version:
            self.content = version["content"]
            self.status = version["status"]
            self._record_version(f"Reverted to version {index}")
            return True
        return False

    def to_dict(self) -> Dict[str, Any]:
        """Convert artifact to dictionary"""
        return {
            "artifact_id": self.artifact_id,
            "artifact_type": self.artifact_type.value,
            "content": self.content,
            "metadata": self.metadata,
            "created_at": self.created_at,
            "updated_at": self.updated_at,
            "status": self.status.name,
            "version_count": len(self.version_history)
        }

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "Artifact":
        """Create an artifact instance from a dictionary"""
        artifact_type = ArtifactType(data["artifact_type"])
        artifact = cls(
            artifact_type=artifact_type,
            content=data["content"],
            metadata=data["metadata"],
            artifact_id=data.get("artifact_id", str(uuid.uuid4()))
        )
        artifact.created_at = data["created_at"]
        artifact.updated_at = data["updated_at"]
        artifact.status = ArtifactStatus[data["status"]]

        # If version history exists, restore it as well
        if "version_history" in data:
            artifact.version_history = data["version_history"]

        return artifact


# coding: utf-8
# Copyright (c) 2025 inclusionAI.
import json
import traceback
import uuid

import aworld.trace as trace

from collections import OrderedDict
from typing import Dict, Any, List, Union, Callable

from aworld.config.conf import AgentConfig, ConfigDict
from aworld.core.agent.agent_desc import get_agent_desc
from aworld.core.agent.base import BaseAgent, AgentResult, is_agent_by_name, is_agent
from aworld.core.common import Observation, ActionModel
from aworld.core.tool.tool_desc import get_tool_desc
from aworld.core.event.base import Message, ToolMessage, Constants
from aworld.logs.util import logger
from aworld.mcp_client.utils import mcp_tool_desc_transform
from aworld.core.memory import MemoryItem
from aworld.memory.main import Memory
from aworld.models.llm import get_llm_model, call_llm_model, acall_llm_model
from aworld.models.model_response import ModelResponse, ToolCall
from aworld.models.utils import tool_desc_transform, agent_desc_transform
from aworld.output import Outputs
from aworld.output.base import StepOutput, MessageOutput
from aworld.sandbox.base import Sandbox
from aworld.utils.common import sync_exec


class Agent(BaseAgent[Observation, List[ActionModel]]):
    """Basic agent for unified protocol within the framework."""

    def __init__(self,
                 conf: Union[Dict[str, Any], ConfigDict, AgentConfig],
                 resp_parse_func: Callable[..., Any] = None,
                 **kwargs):
        """A api class implementation of agent, using the `Observation` and `List[ActionModel]` protocols.

        Args:
            conf: Agent config, supported AgentConfig, ConfigDict or dict.
            resp_parse_func: Response parse function for the agent standard output, transform llm response.
        """
        super(Agent, self).__init__(conf, **kwargs)
        conf = self.conf
        self.model_name = conf.llm_config.llm_model_name if conf.llm_config.llm_model_name else conf.llm_model_name
        self._llm = None
        self.memory = Memory.from_config(
            {"memory_store": kwargs.pop("memory_store") if kwargs.get("memory_store") else "inmemory"})
        self.system_prompt: str = kwargs.pop("system_prompt") if kwargs.get("system_prompt") else conf.system_prompt
        self.agent_prompt: str = kwargs.get("agent_prompt") if kwargs.get("agent_prompt") else conf.agent_prompt

        self.event_driven = kwargs.pop('event_driven', conf.get('event_driven', False))
        self.handler: Callable[..., Any] = kwargs.get('handler')

        self.need_reset = kwargs.get('need_reset') if kwargs.get('need_reset') else conf.need_reset
        # whether to keep contextual information, False means keep, True means reset in every step by the agent call
        self.step_reset = kwargs.get('step_reset') if kwargs.get('step_reset') else True
        # tool_name: [tool_action1, tool_action2, ...]
        self.black_tool_actions: Dict[str, List[str]] = kwargs.get("black_tool_actions") if kwargs.get(
            "black_tool_actions") else conf.get('black_tool_actions', {})
        self.resp_parse_func = resp_parse_func if resp_parse_func else self.response_parse
        self.history_messages = kwargs.get("history_messages") if kwargs.get("history_messages") else 100
        self.use_tools_in_prompt = kwargs.get('use_tools_in_prompt', conf.use_tools_in_prompt)

    def reset(self, options: Dict[str, Any]):
        super().reset(options)
        self.memory = Memory.from_config(
            {"memory_store": options.pop("memory_store") if options.get("memory_store") else "inmemory"})

    @property
    def llm(self):
        # lazy
        if self._llm is None:
            llm_config = self.conf.llm_config or None
            conf = llm_config if llm_config and (
                    llm_config.llm_provider or llm_config.llm_base_url or llm_config.llm_api_key or llm_config.llm_model_name) else self.conf
            self._llm = get_llm_model(conf)
        return self._llm

    def _env_tool(self):
        """Description of agent as tool."""
        return tool_desc_transform(get_tool_desc(),
                                   tools=self.tool_names if self.tool_names else [],
                                   black_tool_actions=self.black_tool_actions)

    def _handoffs_agent_as_tool(self):
        """Description of agent as tool."""
        return agent_desc_transform(get_agent_desc(),
                                    agents=self.handoffs if self.handoffs else [])

    def _mcp_is_tool(self):
        """Description of mcp servers are tools."""
        try:
            return sync_exec(mcp_tool_desc_transform, self.mcp_servers, self.mcp_config)
        except Exception as e:
            logger.error(f"mcp_is_tool error: {e}")
            return []

    def desc_transform(self):
        """Transform of descriptions of supported tools, agents, and MCP servers in the framework to support function calls of LLM."""

        # Stateless tool
        self.tools = self._env_tool()
        # Agents as tool
        self.tools.extend(self._handoffs_agent_as_tool())
        # MCP servers are tools
        self.tools.extend(self._mcp_is_tool())
        return self.tools

    async def async_desc_transform(self):
        """Transform of descriptions of supported tools, agents, and MCP servers in the framework to support function calls of LLM."""

        # Stateless tool
        self.tools = self._env_tool()
        # Agents as tool
        self.tools.extend(self._handoffs_agent_as_tool())
        # MCP servers are tools
        # todo sandbox
        if self.sandbox:
            sand_box = self.sandbox
            mcp_tools = await sand_box.mcpservers.list_tools()
            self.tools.extend(mcp_tools)
        else:
            self.tools.extend(await mcp_tool_desc_transform(self.mcp_servers, self.mcp_config))

    def _messages_transform(
            self,
            observation: Observation,
            sys_prompt: str = None,
            agent_prompt: str = None,
    ):
        messages = []
        if sys_prompt:
            messages.append({'role': 'system', 'content': sys_prompt if self.use_tools_in_prompt else sys_prompt.format(
                tool_list=self.tools)})

        content = observation.content
        if agent_prompt and '{task}' in agent_prompt:
            content = agent_prompt.format(task=observation.content)

        cur_msg = {'role': 'user', 'content': content}
        # query from memory,
        # histories = self.memory.get_last_n(self.history_messages, filter={"session_id": self.context.session_id})
        histories = self.memory.get_last_n(self.history_messages)
        messages.extend(histories)

        action_results = observation.action_result
        if action_results:
            for action_result in action_results:
                cur_msg['role'] = 'tool'
                cur_msg['tool_call_id'] = action_result.tool_id

        agent_info = self.context.context_info.get(self.name())
        if (not self.use_tools_in_prompt and "is_use_tool_prompt" in agent_info and "tool_calls"
                in agent_info and agent_prompt):
            cur_msg['content'] = agent_prompt.format(action_list=agent_info["tool_calls"],
                                                     result=content)

        if observation.images:
            urls = [{'type': 'text', 'text': content}]
            for image_url in observation.images:
                urls.append({'type': 'image_url', 'image_url': {"url": image_url}})

            cur_msg['content'] = urls
        messages.append(cur_msg)
        return messages

    def messages_transform(self,
                           content: str,
                           image_urls: List[str] = None,
                           sys_prompt: str = None,
                           agent_prompt: str = None,
                           **kwargs):
        """Transform the original content to LLM messages of native format.

        Args:
            content: User content.
            image_urls: List of images encoded using base64.
            sys_prompt: Agent system prompt.
            max_step: The maximum list length obtained from memory.
        Returns:
            Message list for LLM.
        """
        messages = []
        if sys_prompt:
            messages.append({'role': 'system', 'content': sys_prompt if self.use_tools_in_prompt else sys_prompt.format(
                tool_list=self.tools)})

        if agent_prompt and '{task}' in agent_prompt:
            content = agent_prompt.format(task=content)

        cur_msg = {'role': 'user', 'content': content}
        # query from memory,
        # histories = self.memory.get_last_n(self.history_messages, filter={"session_id": self.context.session_id})
        histories = self.memory.get_last_n(self.history_messages)
        if histories:
            # default use the first tool call
            for history in histories:
                if self.use_tools_in_prompt and "tool_calls" in history.metadata and history.metadata['tool_calls']:
                    messages.append({'role': history.metadata['role'], 'content': history.content,
                                     'tool_calls': [history.metadata["tool_calls"][0]]})
                else:
                    messages.append({'role': history.metadata['role'], 'content': history.content,
                                     "tool_call_id": history.metadata.get("tool_call_id")})

            if self.use_tools_in_prompt and "tool_calls" in histories[-1].metadata and histories[-1].metadata[
                'tool_calls']:
                tool_id = histories[-1].metadata["tool_calls"][0].id
                if tool_id:
                    cur_msg['role'] = 'tool'
                    cur_msg['tool_call_id'] = tool_id
            if not self.use_tools_in_prompt and "is_use_tool_prompt" in histories[-1].metadata and "tool_calls" in \
                    histories[-1].metadata and agent_prompt:
                cur_msg['content'] = agent_prompt.format(action_list=histories[-1].metadata["tool_calls"],
                                                         result=content)

        if image_urls:
            urls = [{'type': 'text', 'text': content}]
            for image_url in image_urls:
                urls.append({'type': 'image_url', 'image_url': {"url": image_url}})

            cur_msg['content'] = urls
        messages.append(cur_msg)
        return messages

    def use_tool_list(self, resp: ModelResponse) -> List[Dict[str, Any]]:
        tool_list = []
        try:
            if resp and hasattr(resp, 'content') and resp.content:
                content = resp.content.strip()
            else:
                return tool_list
            content = content.replace('\n', '').replace('\r', '')
            response_json = json.loads(content)
            if "use_tool_list" in response_json:
                use_tool_list = response_json["use_tool_list"]
                if use_tool_list:
                    for use_tool in use_tool_list:
                        tool_name = use_tool["tool"]
                        arguments = use_tool["arguments"]
                        if tool_name and arguments:
                            tool_list.append(use_tool)

            return tool_list
        except Exception as e:
            logger.debug(f"tool_parse error, content: {resp.content}, \nerror msg: {e}")
            return tool_list

    def response_parse(self, resp: ModelResponse) -> AgentResult:
        """Default parse response by LLM."""
        results = []
        if not resp:
            logger.warning("LLM no valid response!")
            return AgentResult(actions=[], current_state=None)

        use_tool_list = self.use_tool_list(resp)
        is_call_tool = False
        content = '' if resp.content is None else resp.content
        if resp.tool_calls:
            is_call_tool = True
            for tool_call in resp.tool_calls:
                full_name: str = tool_call.function.name
                if not full_name:
                    logger.warning("tool call response no tool name.")
                    continue
                try:
                    params = json.loads(tool_call.function.arguments)
                except:
                    logger.warning(f"{tool_call.function.arguments} parse to json fail.")
                    params = {}
                # format in framework
                names = full_name.split("__")
                tool_name = names[0]
                if is_agent_by_name(tool_name):
                    param_info = params.get('content', "") + ' ' + params.get('info', '')
                    results.append(ActionModel(tool_name=tool_name,
                                               tool_id=tool_call.id,
                                               agent_name=self.name(),
                                               params=params,
                                               policy_info=content + param_info))
                else:
                    action_name = '__'.join(names[1:]) if len(names) > 1 else ''
                    results.append(ActionModel(tool_name=tool_name,
                                               tool_id=tool_call.id,
                                               action_name=action_name,
                                               agent_name=self.name(),
                                               params=params,
                                               policy_info=content))
        elif use_tool_list and len(use_tool_list) > 0:
            is_call_tool = True
            for use_tool in use_tool_list:
                full_name = use_tool["tool"]
                if not full_name:
                    logger.warning("tool call response no tool name.")
                    continue
                params = use_tool["arguments"]
                if not params:
                    logger.warning("tool call response no tool params.")
                    continue
                names = full_name.split("__")
                tool_name = names[0]
                if is_agent_by_name(tool_name):
                    param_info = params.get('content', "") + ' ' + params.get('info', '')
                    results.append(ActionModel(tool_name=tool_name,
                                               tool_id=use_tool.get('id'),
                                               agent_name=self.name(),
                                               params=params,
                                               policy_info=content + param_info))
                else:
                    action_name = '__'.join(names[1:]) if len(names) > 1 else ''
                    results.append(ActionModel(tool_name=tool_name,
                                               tool_id=use_tool.get('id'),
                                               action_name=action_name,
                                               agent_name=self.name(),
                                               params=params,
                                               policy_info=content))
        else:
            if content:
                content = content.replace("```json", "").replace("```", "")
            # no tool call, agent name is itself.
            results.append(ActionModel(agent_name=self.name(), policy_info=content))
        return AgentResult(actions=results, current_state=None, is_call_tool=is_call_tool)

    def _log_messages(self, messages: List[Dict[str, Any]]) -> None:
        """Log the sequence of messages for debugging purposes"""
        logger.info(f"[agent] Invoking LLM with {len(messages)} messages:")
        for i, msg in enumerate(messages):
            prefix = msg.get('role')
            logger.info(f"[agent] Message {i + 1}: {prefix} ===================================")
            if isinstance(msg['content'], list):
                for item in msg['content']:
                    if item.get('type') == 'text':
                        logger.info(f"[agent] Text content: {item.get('text')}")
                    elif item.get('type') == 'image_url':
                        image_url = item.get('image_url', {}).get('url', '')
                        if image_url.startswith('data:image'):
                            logger.info(f"[agent] Image: [Base64 image data]")
                        else:
                            logger.info(f"[agent] Image URL: {image_url[:30]}...")
            else:
                content = str(msg['content'])
                chunk_size = 500
                for j in range(0, len(content), chunk_size):
                    chunk = content[j:j + chunk_size]
                    if j == 0:
                        logger.info(f"[agent] Content: {chunk}")
                    else:
                        logger.info(f"[agent] Content (continued): {chunk}")

            if 'tool_calls' in msg and msg['tool_calls']:
                for tool_call in msg.get('tool_calls'):
                    if isinstance(tool_call, dict):
                        logger.info(f"[agent] Tool call: {tool_call.get('name')} - ID: {tool_call.get('id')}")
                        args = str(tool_call.get('args', {}))[:1000]
                        logger.info(f"[agent] Tool args: {args}...")
                    elif isinstance(tool_call, ToolCall):
                        logger.info(f"[agent] Tool call: {tool_call.function.name} - ID: {tool_call.id}")
                        args = str(tool_call.function.arguments)[:1000]
                        logger.info(f"[agent] Tool args: {args}...")

    def _agent_result(self, actions: List[ActionModel], caller: str):
        if not actions:
            raise Exception(f'{self.name()} no action decision has been made.')

        tools = OrderedDict()
        agents = []
        for action in actions:
            if is_agent(action):
                agents.append(action)
            else:
                if action.tool_name not in tools:
                    tools[action.tool_name] = []
                tools[action.tool_name].append(action)

        _group_name = None
        # agents and tools exist simultaneously, more than one agent/tool name
        if (agents and tools) or len(agents) > 1 or len(tools) > 1:
            _group_name = f"{self.name()}_{uuid.uuid1().hex}"

        # complex processing
        if _group_name:
            logger.warning(f"more than one agent an tool causing confusion, will choose the first one. {agents}")
            agents = [agents[0]] if agents else []
            for _, v in tools.items():
                actions = v
                break

        if agents:
            return Message(payload=actions,
                           caller=caller,
                           sender=self.name(),
                           receiver=actions[0].tool_name,
                           session_id=self.context.session_id,
                           category=Constants.AGENT)
        else:
            return ToolMessage(payload=actions,
                               caller=caller,
                               sender=self.name(),
                               receiver=actions[0].tool_name,
                               session_id=self.context.session_id)

    def post_run(self, policy_result: List[ActionModel], policy_input: Observation) -> Message:
        return self._agent_result(
            policy_result,
            policy_input.from_agent_name if policy_input.from_agent_name else policy_input.observer
        )

    async def async_post_run(self, policy_result: List[ActionModel], policy_input: Observation) -> Message:
        return self._agent_result(
            policy_result,
            policy_input.from_agent_name if policy_input.from_agent_name else policy_input.observer
        )

    def policy(self, observation: Observation, info: Dict[str, Any] = {}, **kwargs) -> List[ActionModel]:
        """The strategy of an agent can be to decide which tools to use in the environment, or to delegate tasks to other agents.

        Args:
            observation: The state observed from tools in the environment.
            info: Extended information is used to assist the agent to decide a policy.

        Returns:
            ActionModel sequence from agent policy
        """
        output = None
        if kwargs.get("output") and isinstance(kwargs.get("output"), StepOutput):
            output = kwargs["output"]

        # Get current step information for trace recording
        step = kwargs.get("step", 0)
        exp_id = kwargs.get("exp_id", None)
        source_span = trace.get_current_span()

        if hasattr(observation, 'context') and observation.context:
            self.task_histories = observation.context

        self._finished = False
        self.desc_transform()
        images = observation.images if self.conf.use_vision else None
        if self.conf.use_vision and not images and observation.image:
            images = [observation.image]
            observation.images = images
        messages = self.messages_transform(content=observation.content,
                                           image_urls=observation.images,
                                           sys_prompt=self.system_prompt,
                                           agent_prompt=self.agent_prompt)

        self._log_messages(messages)
        self.memory.add(MemoryItem(
            content=messages[-1]['content'],
            metadata={
                "role": messages[-1]['role'],
                "agent_name": self.name(),
                "tool_call_id": messages[-1].get("tool_call_id")
            }
        ))

        llm_response = None
        span_name = f"llm_call_{exp_id}"
        serializable_messages = self._to_serializable(messages)
        with trace.span(span_name) as llm_span:
            llm_span.set_attributes({
                "exp_id": exp_id,
                "step": step,
                "messages": json.dumps(serializable_messages, ensure_ascii=False)
            })
            if source_span:
                source_span.set_attribute("messages", json.dumps(serializable_messages, ensure_ascii=False))

            try:
                llm_response = call_llm_model(
                    self.llm,
                    messages=messages,
                    model=self.model_name,
                    temperature=self.conf.llm_config.llm_temperature,
                    tools=self.tools if self.use_tools_in_prompt and self.tools else None
                )

                logger.info(f"Execute response: {llm_response.message}")
            except Exception as e:
                logger.warn(traceback.format_exc())
                raise e
            finally:
                if llm_response:
                    use_tools = self.use_tool_list(llm_response)
                    is_use_tool_prompt = len(use_tools) > 0
                    if llm_response.error:
                        logger.info(f"llm result error: {llm_response.error}")
                    else:
                        info = {
                            "role": "assistant",
                            "agent_name": self.name(),
                            "tool_calls": llm_response.tool_calls if self.use_tools_in_prompt else use_tools,
                            "is_use_tool_prompt": is_use_tool_prompt if self.use_tools_in_prompt else False
                        }
                        self.memory.add(MemoryItem(
                            content=llm_response.content,
                            metadata=info
                        ))
                        # rewrite
                        self.context.context_info[self.name()] = info
                else:
                    logger.error(f"{self.name()} failed to get LLM response")
                    raise RuntimeError(f"{self.name()} failed to get LLM response")

        agent_result = sync_exec(self.resp_parse_func, llm_response)
        if not agent_result.is_call_tool:
            self._finished = True

        if output:
            output.add_part(MessageOutput(source=llm_response, json_parse=False))
            output.mark_finished()
        return agent_result.actions

    async def async_policy(self, observation: Observation, info: Dict[str, Any] = {}, **kwargs) -> List[ActionModel]:
        """The strategy of an agent can be to decide which tools to use in the environment, or to delegate tasks to other agents.

        Args:
            observation: The state observed from tools in the environment.
            info: Extended information is used to assist the agent to decide a policy.

        Returns:
            ActionModel sequence from agent policy
        """
        outputs = None
        if kwargs.get("outputs") and isinstance(kwargs.get("outputs"), Outputs):
            outputs = kwargs.get("outputs")

        # Get current step information for trace recording
        step = kwargs.get("step", 0)
        exp_id = kwargs.get("exp_id", None)
        source_span = trace.get_current_span()

        if hasattr(observation, 'context') and observation.context:
            self.task_histories = observation.context

        self._finished = False
        await self.async_desc_transform()
        images = observation.images if self.conf.use_vision else None
        if self.conf.use_vision and not images and observation.image:
            images = [observation.image]
        messages = self.messages_transform(content=observation.content,
                                           image_urls=images,
                                           sys_prompt=self.system_prompt,
                                           agent_prompt=self.agent_prompt)

        self._log_messages(messages)
        self.memory.add(MemoryItem(
            content=messages[-1]['content'],
            metadata={
                "role": messages[-1]['role'],
                "agent_name": self.name(),
                "tool_call_id": messages[-1].get("tool_call_id")
            }
        ))

        llm_response = None
        span_name = f"llm_call_{exp_id}"
        serializable_messages = self._to_serializable(messages)
        with trace.span(span_name) as llm_span:
            llm_span.set_attributes({
                "exp_id": exp_id,
                "step": step,
                "messages": json.dumps(serializable_messages, ensure_ascii=False)
            })
            if source_span:
                source_span.set_attribute("messages", json.dumps(serializable_messages, ensure_ascii=False))

            try:
                llm_response = await acall_llm_model(
                    self.llm,
                    messages=messages,
                    model=self.model_name,
                    temperature=self.conf.llm_config.llm_temperature,
                    tools=self.tools if self.use_tools_in_prompt and self.tools else None,
                    stream=kwargs.get("stream", False)
                )
                if outputs and isinstance(outputs, Outputs):
                    await outputs.add_output(MessageOutput(source=llm_response, json_parse=False))

                # Record LLM response
                llm_span.set_attributes({
                    "llm_response": json.dumps(llm_response.to_dict(), ensure_ascii=False),
                    "tool_calls": json.dumps([tool_call.model_dump() for tool_call in
                                              llm_response.tool_calls] if llm_response.tool_calls else [],
                                             ensure_ascii=False),
                    "error": llm_response.error if llm_response.error else ""
                })

            except Exception as e:
                logger.warn(traceback.format_exc())
                llm_span.set_attribute("error", str(e))
                raise e
            finally:
                if llm_response:
                    use_tools = self.use_tool_list(llm_response)
                    is_use_tool_prompt = len(use_tools) > 0
                    if llm_response.error:
                        logger.info(f"llm result error: {llm_response.error}")
                    else:
                        self.memory.add(MemoryItem(
                            content=llm_response.content,
                            metadata={
                                "role": "assistant",
                                "agent_name": self.name(),
                                "tool_calls": llm_response.tool_calls if self.use_tools_in_prompt else use_tools,
                                "is_use_tool_prompt": is_use_tool_prompt if self.use_tools_in_prompt else False
                            }
                        ))
                else:
                    logger.error(f"{self.name()} failed to get LLM response")
                    raise RuntimeError(f"{self.name()} failed to get LLM response")

        agent_result = sync_exec(self.resp_parse_func, llm_response)
        if not agent_result.is_call_tool:
            self._finished = True
        return agent_result.actions

    def _to_serializable(self, obj):
        if isinstance(obj, dict):
            return {k: self._to_serializable(v) for k, v in obj.items()}
        elif isinstance(obj, list):
            return [self._to_serializable(i) for i in obj]
        elif hasattr(obj, "to_dict"):
            return obj.to_dict()
        elif hasattr(obj, "model_dump"):
            return obj.model_dump()
        elif hasattr(obj, "dict"):
            return obj.dict()
        else:
            return obj



import time
import sys
import os
from logging import StreamHandler
from abc import ABC
from logging import Logger, NOTSET, LogRecord, Filter, Formatter, Handler
from typing import Optional, Union

from aworld.trace.base import get_tracer_provider_silent, Tracer, AttributeValueType

TRACE_LOG_FORMAT = '%(asctime)s - [%(trace_id)s] - [%(span_id)s] - %(name)s - %(levelname)s - %(message)s'
SPECIAL_TRACE_LOG_FORMAT = '%(asctime)s - [trace_%(trace_id)s] - [%(span_id)s] - %(name)s - %(levelname)s - %(message)s'


class LoggerProvider(ABC):
    """A logger provider is a factory for loggers."""


_GLOBAL_LOG_PROVIDER: Optional[LoggerProvider] = None


def set_log_provider(provider: str = "otlp",
                     backend: str = "logfire",
                     base_url: str = None,
                     write_token: str = None,
                     **kwargs):
    """Set the global log provider."""

    global _GLOBAL_LOG_PROVIDER

    if provider == "otlp":
        from .opentelemetry.otlp_log import OTLPLoggerProvider
        _GLOBAL_LOG_PROVIDER = OTLPLoggerProvider(backend=backend,
                                                  base_url=base_url,
                                                  write_token=write_token,
                                                  **kwargs)


def get_log_provider() -> LoggerProvider:
    """
    Get the global log provider.
    """
    global _GLOBAL_LOG_PROVIDER
    if _GLOBAL_LOG_PROVIDER is None:
        raise ValueError("No log provider has been set.")
    return _GLOBAL_LOG_PROVIDER


def instrument_logging(logger: Logger, level: Union[int, str] = NOTSET) -> None:
    """Instrument the logger."""
    for handler in logger.root.handlers:
        if not any(isinstance(filter, TraceLoggingFilter) for filter in handler.filters):
            handler.setFormatter(Formatter(TRACE_LOG_FORMAT))
            handler.addFilter(TraceLoggingFilter())

    if not logger.handlers:
        print("No handlers found, adding a StreamHandler. logger=", logger.name)
        handler = StreamHandler()
        handler.setFormatter(Formatter(SPECIAL_TRACE_LOG_FORMAT))
        handler.addFilter(TraceLoggingFilter())
        logger.addHandler(handler)
    else:
        for handler in logger.handlers:
            if not any(isinstance(filter, TraceLoggingFilter) for filter in handler.filters):
                handler.setFormatter(Formatter(SPECIAL_TRACE_LOG_FORMAT))
                handler.addFilter(TraceLoggingFilter())
    logger.propagate = False
    logger.addHandler(TraceLogginHandler(level))


class TraceLoggingFilter(Filter):
    """
    A filter that adds trace information to log records.
    """

    def filter(self, record: LogRecord) -> bool:
        """
        Add trace information to the log record.
        """
        trace = get_tracer_provider_silent()
        if trace:
            span = trace.get_current_span()
            record.trace_id = span.get_trace_id() if span else None
            record.span_id = span.get_span_id() if span else None
        return True


class TraceLogginHandler(Handler):
    """
    A handler class which writes logging records, appropriately formatted,
    to a stream. Note that this class does not close the stream, as
    sys.stdout or sys.stderr may be used.
    """
    @staticmethod
    def strip_color(text: str) -> str:
        """Remove ANSI color codes from text"""
        import re
        return re.sub(r'\033\[[0-9;]*m', '', text)

    def __init__(self,
                 level: Union[int, str] = NOTSET,
                 tracer_name: str = "aworld.log") -> None:
        """Initialize the handler."""
        super().__init__(level=level)
        self._tracer_name = tracer_name
        self._tracer: Tracer = None

    def emit(self, record: LogRecord) -> None:
        """Emit a record."""
        trace = get_tracer_provider_silent()
        if not trace or not trace.get_current_span() or not trace.get_current_span().is_recording():
            return

        if not self._tracer:
            self._tracer = trace.get_tracer(name=self._tracer_name)

        try:
            f = sys._getframe()
            while f:
                if 'logging/__init__.py' in f.f_code.co_filename or \
                        f.f_code.co_filename.startswith(os.path.dirname(__file__)):
                    f = f.f_back
                else:
                    break

            origin_msg = record.msg
            raw_msg = None
            if f:
                try:
                    import linecache
                    line = linecache.getline(f.f_code.co_filename, f.f_lineno)
                    if 'logger.' in line:
                        raw_msg = line.split('logger.', 1)[1].split(
                            '(', 1)[1].split(')', 1)[0].strip()
                except:
                    pass
            record.msg = self.strip_color(record.msg)
            msg_template = raw_msg if raw_msg else record.msg

            if len(msg_template) > 255:
                msg_template = msg_template[:255] + '...'

            attributes = {
                'code.filepath': f.f_code.co_filename if f else record.pathname,
                'code.lineno': f.f_lineno if f else record.lineno,
                'code.function': f.f_code.co_name if f else record.funcName,
                'log.level': record.levelname,
                'log.logger': record.name,
                'log.message': self.format(record),
            }
            record.msg = origin_msg
            self._create_span(
                span_name=msg_template,
                attributes=attributes,
                exc_info=record.exc_info,
            )
        except RecursionError:  # See issue 36272
            raise
        except Exception:
            self.handleError(record)

    def _create_span(self,
                     span_name: str,
                     attributes: dict[str, AttributeValueType] = None,
                     exc_info: BaseException = None):
        start_time = time.time_ns()
        span = self._tracer.start_span(
            name=span_name,
            attributes=attributes,
            start_time=start_time,
        )
        if exc_info:
            span.record_exception(exception=exc_info, timestamp=start_time)
        span.end()



import asyncio

from aworld.sandbox import Sandbox, SandboxEnvType


# Define an asynchronous function to call asynchronous methods
async def run_async_tasks(sand_box):
    tools = await sand_box.mcpservers.list_tools()
    print(f"Tools: {tools}")
    return tools

if __name__ == "__main__":
    # 只使用memory服务器
    mcp_servers = ["memory","amap-amap-sse"]
    mcp_config = {
      "mcpServers": {
        "memory": {
          "type": "stdio",
          "command": "npx",
          "args": [
            "-y",
            "@modelcontextprotocol/server-memory"
          ]
        }

      }
    }
    sand_box = Sandbox(mcp_servers=mcp_servers, mcp_config=mcp_config,env_type=SandboxEnvType.SUPERCOMPUTER)
    print(f"Sandbox ID: {sand_box.sandbox_id}")
    print(f"Status: {sand_box.status}")
    print(f"Timeout: {sand_box.timeout}")
    print(f"Metadata: {sand_box.metadata}")
    print(f"Environment Type: {sand_box.env_type}")
    print(f"MCP Servers: {sand_box.mcp_servers}")
    print(f"MCP Config: {sand_box.mcp_config}")
    
    # Use asyncio to run asynchronous methods
    asyncio.run(run_async_tasks(sand_box))
    
    print(f"MCP Servers-new: {sand_box.mcp_servers}")
    print(f"MCP Config-new: {sand_box.mcp_config}")



from __future__ import annotations

import ast
import uuid
import time
from pathlib import Path
from collections import deque
from functools import partial
from typing import TYPE_CHECKING, Any, Callable, ContextManager, cast

from aworld.trace.base import AttributeValueType
from aworld.trace.constants import ATTRIBUTES_MESSAGE_TEMPLATE_KEY

if TYPE_CHECKING:
    from .context_manager import TraceManager
    from .auto_trace import not_auto_trace


def compile_source(
        tree: ast.AST, filename: str, module_name: str, trace_manager: TraceManager, min_duration_ns: int
) -> Callable[[dict[str, Any]], None]:
    """Compile a modified AST of the module's source code in the module's namespace.

    Returns a function which accepts module globals and executes the compiled code.

    The modified AST wraps the body of every function definition in `with context_factories[index]():`.
    `context_factories` is added to the module's namespace as `aworld_<uuid>`.
    `index` is a different constant number for each function definition.
    """

    context_factories_var_name = f'aworld_{uuid.uuid4().hex}'
    # The variable name for storing context_factors in the module's namespace.

    context_factories: list[Callable[[], ContextManager[Any]]] = []
    tree = rewrite_ast(tree, filename, context_factories_var_name, module_name, trace_manager, context_factories,
                       min_duration_ns)
    assert isinstance(tree, ast.Module)  # for type checking
    # dont_inherit=True is necessary to prevent the module from inheriting the __future__ import from this module.
    code = compile(tree, filename, 'exec', dont_inherit=True)

    def execute(globs: dict[str, Any]):
        globs[context_factories_var_name] = context_factories
        exec(code, globs, globs)

    return execute


def rewrite_ast(
        tree: ast.AST,
        filename: str,
        context_factories_var_name: str,
        module_name: str,
        trace_manager: TraceManager,
        context_factories: list[Callable[[], ContextManager[Any]]],
        min_duration_ns: int,
) -> ast.AST:
    transformer = AutoTraceTransformer(
        context_factories_var_name, filename, module_name, trace_manager, context_factories, min_duration_ns
    )
    return transformer.visit(tree)


class AutoTraceTransformer(ast.NodeTransformer):
    """Trace all encountered functions except those explicitly marked with `@no_auto_trace`."""

    def __init__(
            self,
            context_factories_var_name: str,
            filename: str,
            module_name: str,
            trace_manager: TraceManager,
            context_factories: list[Callable[[], ContextManager[Any]]],
            min_duration_ns: int,
    ):
        self._context_factories_var_name = context_factories_var_name
        self._filename = filename
        self._module_name = module_name
        self._trace_manager = trace_manager
        self._context_factories = context_factories
        self._min_duration_ns = min_duration_ns
        self._qualname_stack: list[str] = []

    def visit_ClassDef(self, node: ast.ClassDef):
        """Visit a class definition and rewrite its methods."""

        if self.check_not_auto_trace(node):
            return node

        self._qualname_stack.append(node.name)
        node = cast(ast.ClassDef, self.generic_visit(node))
        self._qualname_stack.pop()
        return node

    def visit_FunctionDef(self, node: ast.FunctionDef) -> ast.AST:
        """Visit a function definition and rewrite it."""

        if self.check_not_auto_trace(node):
            return node

        self._qualname_stack.append(node.name)
        qualname = '.'.join(self._qualname_stack)
        self._qualname_stack.append('<locals>')
        self.generic_visit(node)
        self._qualname_stack.pop()  # <locals>
        self._qualname_stack.pop()  # node.name
        return self.rewrite_function(node, qualname)

    def check_not_auto_trace(self, node: ast.FunctionDef | ast.AsyncFunctionDef | ast.ClassDef) -> bool:
        """Return true if the node has a `@not_auto_trace` decorator."""
        return any(
            (
                    isinstance(node, ast.Name)
                    and node.id == not_auto_trace.__name__
                # or (
                #     isinstance(node, ast.Attribute)
                #     and node.attr == not_auto_trace.__name__
                #     and isinstance(node.value, ast.Name)
                #     and node.value.id == xxx.__name__
                # )
            )
            for node in node.decorator_list
        )

    def rewrite_function(self, node: ast.FunctionDef | ast.AsyncFunctionDef, qualname: str) -> ast.AST:
        """Rewrite a function definition to trace its execution."""

        if has_yield(node):
            return node

        body = node.body.copy()
        new_body: list[ast.stmt] = []
        if (
                body
                and isinstance(body[0], ast.Expr)
                and isinstance(body[0].value, ast.Constant)
                and isinstance(body[0].value.value, str)
        ):
            new_body.append(body.pop(0))

        if not body or (
                len(body) == 1
                and (
                        isinstance(body[0], ast.Pass)
                        or (isinstance(body[0], ast.Expr) and isinstance(body[0].value, ast.Constant))
                )
        ):
            return node

        span = ast.With(
            items=[
                ast.withitem(
                    context_expr=self.trace_context_method_call_node(node, qualname),
                )
            ],
            body=body,
            type_comment=node.type_comment,
        )
        new_body.append(span)

        return ast.fix_missing_locations(
            ast.copy_location(
                type(node)(  # type: ignore
                    name=node.name,
                    args=node.args,
                    body=new_body,
                    decorator_list=node.decorator_list,
                    returns=node.returns,
                    type_comment=node.type_comment,
                ),
                node,
            )
        )

    def trace_context_method_call_node(self, node: ast.FunctionDef | ast.AsyncFunctionDef, qualname: str) -> ast.Call:
        """Return a method call to `context_factories[index]()`."""

        index = len(self._context_factories)
        span_factory = partial(
            self._trace_manager._create_auto_span,  # type: ignore
            *self.build_create_auto_span_args(qualname, node.lineno),
        )
        if self._min_duration_ns > 0:

            timer = time.time_ns
            min_duration = self._min_duration_ns

            # This needs to be as fast as possible since it's the cost of auto-tracing a function
            # that never actually gets instrumented because its calls are all faster than `min_duration`.
            class MeasureTime:
                __slots__ = 'start'

                def __enter__(_self):
                    _self.start = timer()

                def __exit__(_self, *_):
                    # the first call exceeding min_ruration will not be tracked, and subsequent calls will only be tracked
                    if timer() - _self.start >= min_duration:
                        self._context_factories[index] = span_factory

            self._context_factories.append(MeasureTime)
        else:
            self._context_factories.append(span_factory)

        # This node means:
        #   context_factories[index]()
        # where `context_factories` is a global variable with the name `self._context_factories_var_name`
        # pointing to the `self.context_factories` list.
        return ast.Call(
            func=ast.Subscript(
                value=ast.Name(id=self._context_factories_var_name, ctx=ast.Load()),
                slice=ast.Index(value=ast.Constant(value=index)),  # type: ignore
                ctx=ast.Load(),
            ),
            args=[],
            keywords=[],
        )

    def build_create_auto_span_args(self, qualname: str, lineno: int) -> tuple[str, dict[str, AttributeValueType]]:
        """Build the arguments for `create_auto_span`."""

        stack_info = {
            'code.filepath': get_filepath(self._filename),
            'code.lineno': lineno,
            'code.function': qualname,
        }
        attributes: dict[str, AttributeValueType] = {**stack_info}  # type: ignore

        msg_template = f'Calling {self._module_name}.{qualname}'
        attributes[ATTRIBUTES_MESSAGE_TEMPLATE_KEY] = msg_template

        span_name = msg_template

        return span_name, attributes


def has_yield(node: ast.AST):
    """Return true if the node has a yield statement."""

    queue = deque([node])
    while queue:
        node = queue.popleft()
        for child in ast.iter_child_nodes(node):
            if isinstance(child, (ast.Yield, ast.YieldFrom)):
                return True
            if not isinstance(child, (ast.FunctionDef, ast.AsyncFunctionDef, ast.Lambda)):
                queue.append(child)


def get_filepath(file: str):
    """Return a dict with the filepath attribute."""

    path = Path(file)
    if path.is_absolute():
        try:
            path = path.relative_to(Path('.').resolve())
        except ValueError:  # pragma: no cover
            # happens if filename path is not within CWD
            pass
    return str(path)



"""
Image MCP Server

This module provides MCP server functionality for image processing and analysis.
It handles image encoding, optimization, and various image analysis tasks such as
OCR (Optical Character Recognition) and visual reasoning.

The server supports both local image files and remote image URLs with proper validation
and handles various image formats including JPEG, PNG, GIF, and others.

Main functions:
- encode_images: Encodes images to base64 format with optimization
- optimize_image: Resizes and optimizes images for better performance
- Various MCP tools for image analysis and processing
"""

# import asyncio
import base64
import os
from io import BytesIO
from typing import Any, Dict, List

from mcp.server.fastmcp import FastMCP
from openai import OpenAI
from PIL import Image
from pydantic import Field

from aworld.logs.util import logger
from mcp_servers.utils import get_file_from_source

# Initialize MCP server
mcp = FastMCP("image-server")


IMAGE_OCR = (
    "Input is a base64 encoded image. Read text from image if present. "
    "Return a json string with the following format: "
    '{"image_text": "text from image"}'
)

IMAGE_REASONING = (
    "Input is a base64 encoded image. Given user's task: {task}, "
    "solve it following the guide line:\n"
    "1. Careful visual inspection\n"
    "2. Contextual reasoning\n"
    "3. Text transcription where relevant\n"
    "4. Logical deduction from visual evidence\n"
    "Return a json string with the following format: "
    '{"image_reasoning_result": "reasoning result given task and image"}'
)


def optimize_image(image_data: bytes, max_size: int = 1024) -> bytes:
    """
    Optimize image by resizing if needed

    Args:
        image_data: Raw image data
        max_size: Maximum dimension size in pixels

    Returns:
        bytes: Optimized image data

    Raises:
        ValueError: When image cannot be processed
    """
    try:
        image = Image.open(BytesIO(image_data))

        # Resize if image is too large
        if max(image.size) > max_size:
            ratio = max_size / max(image.size)
            new_size = (int(image.size[0] * ratio), int(image.size[1] * ratio))
            image = image.resize(new_size, Image.Resampling.LANCZOS)

        # Save to buffer
        buffered = BytesIO()
        image_format = image.format if image.format else "JPEG"
        image.save(buffered, format=image_format)
        return buffered.getvalue()

    except Exception as e:
        logger.warning(f"Failed to optimize image: {str(e)}")
        return image_data  # Return original data if optimization fails


def encode_images(image_sources: List[str], with_header: bool = True) -> List[str]:
    """
    Encode images to base64 format with robust file handling

    Args:
        image_sources: List of URLs or local file paths of images
        with_header: Whether to include MIME type header

    Returns:
        List[str]: Base64 encoded image strings, with MIME type prefix if with_header is True

    Raises:
        ValueError: When image source is invalid or image format is not supported
    """
    if not image_sources:
        raise ValueError("Image sources cannot be empty")

    images = []
    for image_source in image_sources:
        try:
            # Get file with validation (only image files allowed)
            file_path, mime_type, content = get_file_from_source(
                image_source,
                allowed_mime_prefixes=["image/"],
                max_size_mb=10.0,  # 10MB limit for images
                type="image",
            )

            # Optimize image
            optimized_content = optimize_image(content)

            # Encode to base64
            image_base64 = base64.b64encode(optimized_content).decode()

            # Format with header if requested
            final_image = f"data:{mime_type};base64,{image_base64}" if with_header else image_base64

            images.append(final_image)

            # Clean up temporary file if it was created for a URL
            if file_path != os.path.abspath(image_source) and os.path.exists(file_path):
                os.unlink(file_path)

        except Exception as e:
            logger.error(f"Error encoding image from {image_source}: {str(e)}")
            raise

    return images


def image_to_base64(image_path):
    try:
        # 打开图片
        with Image.open(image_path) as image:
            buffered = BytesIO()
            image_format = image.format if image.format else "JPEG"
            image.save(buffered, format=image_format)
            image_bytes = buffered.getvalue()
            base64_encoded = base64.b64encode(image_bytes).decode("utf-8")
            return base64_encoded
    except Exception as e:
        print(f"Base64 error: {e}")
        return None


def create_image_contents(prompt: str, image_base64: List[str]) -> List[Dict[str, Any]]:
    """Create uniform image format for querying llm."""
    content = [
        {"type": "text", "text": prompt},
    ]
    content.extend([{"type": "image_url", "image_url": {"url": url}} for url in image_base64])
    return content


@mcp.tool(
    description="solve the question by careful reasoning given the image(s) in given filepath or url, including reasoning, ocr, etc."
)
def mcp_image_recognition(
    image_urls: List[str] = Field(description="The input image(s) in given a list of filepaths or urls."),
    question: str = Field(description="The question to ask."),
) -> str:
    """solve the question by careful reasoning given the image(s) in given filepath or url."""

    try:
        image_base64 = image_to_base64(image_urls[0])
        logger.info(f"image_base64:{image_urls[0]}")
        reasoning_prompt = question
        messages = [
            {"role": "system", "content": "You are a helpful assistant."},
            {
                "role": "user",
                "content": [
                    {"type": "text", "text": reasoning_prompt},
                    {
                        "type": "image_url",
                        "image_url": {"url": f"data:image/jpeg;base64,{image_base64}"},
                    },
                ],
            },
        ]

        client = OpenAI(api_key=os.getenv("IMAGE_LLM_API_KEY"), base_url=os.getenv("IMAGE_LLM_BASE_URL"))
        response = client.chat.completions.create(
            model=os.getenv("IMAGE_LLM_MODEL_NAME"),
            messages=messages,
        )

        logger.info(f"response:{response.choices[0]}")
        image_reasoning_result = response.choices[0].message.content

    except Exception as e:
        image_reasoning_result = ""
        import traceback

        traceback.print_exc()
        logger.error(f"image_reasoning_result-Execute error: {e}")

    logger.info(f"---get_reasoning_by_image-image_reasoning_result:{image_reasoning_result}")

    return image_reasoning_result


def main():
    from dotenv import load_dotenv

    load_dotenv()

    print("Starting Image MCP Server...", file=sys.stderr)
    mcp.run(transport="stdio")


# Make the module callable
def __call__():
    """
    Make the module callable for uvx.
    This function is called when the module is executed directly.
    """
    main()


# Add this for compatibility with uvx
import sys

sys.modules[__name__].__call__ = __call__

# Run the server when the script is executed directly
if __name__ == "__main__":
    main()



# coding: utf-8
# Copyright (c) 2025 inclusionAI.
import json
from typing import List, Tuple, Dict, Any

from aworld.core.tool.base import AsyncTool
from aworld.core.common import Observation, ActionModel, Config
from aworld.logs.util import logger
from aworld.tools.utils import build_observation


class TemplateTool(AsyncTool):
    def __init__(self, conf: Config, **kwargs) -> None:
        super(TemplateTool, self).__init__(conf, **kwargs)

    async def reset(self, *, seed: int | None = None, options: Dict[str, str] | None = None) -> Tuple[
        Observation, dict[str, Any]]:
        # from options obtain user query
        return build_observation(observer=self.name(),
                                 ability='',
                                 content=options.get("query", None) if options else None), {}

    async def do_step(self,
                      action: List[ActionModel],
                      **kwargs) -> Tuple[Observation, float, bool, bool, Dict[str, Any]]:
        reward = 0
        fail_error = ""
        action_result = None

        invalid_acts: List[int] = []
        for i, act in enumerate(action):
            if act.tool_name != self.name():
                logger.warning(f"tool {act.tool_name} is not a {self.name()} tool!")
                invalid_acts.append(i)

        if invalid_acts:
            for i in invalid_acts:
                action[i] = None

        resp = ""
        try:
            action_result, resp = await self.action_executor.async_execute_action(action, **kwargs)
            reward = 1
        except Exception as e:
            fail_error = str(e)

        terminated = kwargs.get("terminated", False)
        for res in action_result:
            if res.is_done:
                terminated = res.is_done
                self._finished = True

        info = {"exception": fail_error}
        info.update(kwargs)
        if resp:
            resp = json.dumps(resp)
        else:
            resp = action_result[0].content

        observation = build_observation(observer=self.name(),
                                        action_result=action_result,
                                        ability=action[-1].action_name,
                                        content=resp)
        return (observation,
                reward,
                terminated,
                kwargs.get("truncated", False),
                info)

    async def close(self) -> None:
        pass

    async def finished(self) -> bool:
        # one time
        return True



import hashlib
import json
import time
import uuid
from enum import Enum
from pathlib import Path
from typing import Dict, Any, Optional, List, Literal

from pydantic import BaseModel


class ArtifactRepository:
    def __init__(self):
        """
        Initialize the artifact repository
        """
        pass

    def _load_index(self) -> Dict[str, Any]:
        """Load or create index file"""
        pass

    def _save_index(self, index: Dict[str, Any]) -> None:
        """Save index to file"""
        pass

    def _compute_content_hash(self, data: Any) -> str:
        """
        Calculate the hash value of the content as a storage identifier

        Args:
            data: Data to be stored

        Returns:
            SHA-256 hash value of the content
        """
        pass

    def store(self,
              artifact_id: str,
              type: Literal['artifact', 'workspace'],
              data: Dict[str, Any],
              metadata: Optional[Dict[str, Any]] = None
              ) -> str:
        """
        Store artifact and return its version identifier

        Args:
            artifact_id: Unique identifier of the artifact
            data: Data to be stored
            metadata: Optional metadata

        Returns:
            Version identifier
        """
        pass

    def retrieve(self, version_id: str) -> Optional[Dict[str, Any]]:
        """
        Retrieve artifact based on version ID

        Args:
            version_id: Version identifier

        Returns:
            Stored data, or None if it doesn't exist
        """
        pass

    def retrieve_latest_artifact(self, artifact_id: str) -> Optional[Dict[str, Any]]:
        pass

    def get_versions(self, artifact_id: str) -> List[Dict[str, Any]]:
        """
        Get information about all versions of an artifact

        Args:
            artifact_id: Artifact identifier

        Returns:
            List of version information
        """
        pass



class CommonEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, Enum):
            return obj.name
        if isinstance(obj, BaseModel):
            return obj.model_dump()
        return json.JSONEncoder.default(self, obj)

class EnumDecoder(json.JSONDecoder):
    def decode(self, s, **kwargs):
        parsed_json = super().decode(s, **kwargs)
        for key, value in parsed_json.items():
            if isinstance(value, dict) and value.get("__enum__"):
                enum_type = globals()[value["__enum_type__"]]
                enum_value = enum_type[value["__enum_value__"]]
                parsed_json[key] = enum_value
        return parsed_json


class LocalArtifactRepository(ArtifactRepository):
    """Artifact storage layer: manages versioned artifacts through content-addressable storage"""

    def __init__(self, storage_path: str):
        """
        Initialize the artifact repository
        
        Args:
            storage_path: Directory path for storing data
        """
        super().__init__()
        self.storage_path = Path(storage_path)
        self.storage_path.mkdir(parents=True, exist_ok=True)
        self.index_path = self.storage_path / "index.json"
        self.index = self._load_index()

    def _load_index(self) -> Dict[str, Any]:
        """Load or create index file"""
        if self.index_path.exists():
            try:
                with open(self.index_path, 'r') as f:
                    return json.load(f)
            except json.JSONDecodeError:
                return {"artifacts": [], "versions": []}
        else:
            index = {"artifacts": [], "versions": []}
            self._save_index(index)
            return index

    def _save_index(self, index: Dict[str, Any]) -> None:
        """Save index to file"""
        with open(self.index_path, 'w') as f:
            json.dump(index, f, indent=2, ensure_ascii=False, cls=CommonEncoder)

    def _compute_content_hash(self, data: Any) -> str:
        """
        Calculate the hash value of the content as a storage identifier
        
        Args:
            data: Data to be stored
            
        Returns:
            SHA-256 hash value of the content
        """
        content = json.dumps(data, sort_keys=True, cls=CommonEncoder).encode('utf-8')
        return hashlib.sha256(content).hexdigest()

    def store(self,
              artifact_id: str,
              type: str,
              data: Dict[str, Any],
              metadata: Optional[Dict[str, Any]] = None
              ) -> str:
        """
        Store artifact and return its version identifier
        
        Args:
            artifact_id: Unique identifier of the artifact
            data: Data to be stored
            metadata: Optional metadata
            
        Returns:
            Version identifier
        """
        # Calculate content hash
        content_hash = self._compute_content_hash(data)

        # Create version record
        version = {
            "hash": content_hash,
            "timestamp": time.time(),
            "metadata": metadata or {}
        }

        # Store content
        content_path = self.storage_path / f"{type}_{content_hash}.json"
        if not content_path.exists():
            with open(content_path, 'w') as f:
                json.dump(data, f, indent=2, ensure_ascii=False, cls=CommonEncoder)

        # Update index
        if type == 'artifact':
            if artifact_id not in self.index["artifacts"]:
                self.index["artifacts"].append({
                    'artifact_id': artifact_id,
                    'type': type,
                    'version': version
                })
        elif type == 'workspace':
            version['version_id'] = str(uuid.uuid4())

            self.index["versions"].append(
                version
            )
        self._save_index(self.index)

        return "success"

    def retrieve(self, version_id: str) -> Optional[Dict[str, Any]]:
        """
        Retrieve artifact based on version ID
        
        Args:
            version_id: Version identifier
            
        Returns:
            Stored data, or None if it doesn't exist
        """
        for version in self.index["versions"]:
            if version_id != version['version_id']:
                continue
            content_hash = version["hash"]
            content_path = self.storage_path / f"workspace_{content_hash}.json"

            if not content_path.exists():
                return None

            with open(content_path, 'r') as f:
                return json.load(f)
        return None

    def retrieve_latest_artifact(self, artifact_id: str) -> Optional[Dict[str, Any]]:
        """
        Retrieve artifact based on version ID

        Args:
            version_id: Version identifier

        Returns:
            Stored data, or None if it doesn't exist
        """
        for artifact in self.index["artifacts"]:
            if artifact['artifact_id'] != artifact_id:
                continue
            content_hash = artifact["version"]["hash"]
            content_path = self.storage_path / f"artifact_{content_hash}.json"

            if not content_path.exists():
                return None

            with open(content_path, 'r') as f:
                return json.load(f)
        return None

    def get_versions(self, artifact_id: str) -> List[Dict[str, Any]]:
        """
        Get information about all versions of an artifact
        
        Args:
            artifact_id: Artifact identifier
            
        Returns:
            List of version information
        """
        if artifact_id not in self.index["artifacts"]:
            return []

        versions = []
        for version_id in self.index["artifacts"][artifact_id]:
            version_info = self.index["versions"][version_id].copy()
            version_info["id"] = version_id
            versions.append(version_info)

        return versions



# coding: utf-8
# Copyright (c) 2025 inclusionAI.

from typing import List, Dict

from aworld.core.tool.action_factory import ActionFactory

from aworld.core.tool.base import ToolFactory
from aworld.logs.util import logger


def tool_action_desc():
    """Utility method of generate description of tools and their actions.

    The standard protocol can be transformed based on the API of different llm.
    Define as follows:
    ```
    {
        "tool_name": {
            "desc": "A toolkit description.",
            "actions": [
                {
                    "name": "action name",
                    "desc": "action description.",
                    "params": {
                        "param_name": {
                            "desc": "param description.",
                            "type": "param type, such as int, str, etc.",
                            "required": True | False
                        }
                    }
                }
            ]
        }
    }
    ```
    """

    def process(action_info):
        action_dict = dict()
        action_dict["name"] = action_info.name
        action_dict["desc"] = action_info.desc
        action_dict["params"] = dict()

        for k, v in action_info.input_params.items():
            params_dict = v.model_dump()
            params_dict.pop("name")
            action_dict["params"][k] = params_dict
        return action_dict

    descs = dict()
    for tool in ToolFactory:
        tool_val_dict = dict()
        descs[tool] = tool_val_dict

        tool_val_dict["desc"] = ToolFactory.desc(tool)
        tool_action = ToolFactory.get_tool_action(tool)
        actions = []
        action_names = ActionFactory.get_actions_by_tool(tool_name=tool)
        if action_names:
            for action_name in action_names:
                info = tool_action.get_value_by_name(action_name)
                if not info:
                    logger.warning(f"{action_name} can not find in {tool}, please check it.")
                    continue
                try:
                    action_dict = process(info)
                except:
                    logger.warning(f"{action_name} process fail.")
                    action_dict = dict()
                actions.append(action_dict)
        elif tool_action:
            for k, info in tool_action.__members__.items():
                action_dict = process(info.value)
                actions.append(action_dict)
        else:
            logger.debug(f"{tool} no action!")
        tool_val_dict["actions"] = actions
    return descs


def get_actions() -> List[str]:
    res = []
    for _, tool_info in tool_action_desc().items():
        actions = tool_info.get("actions")
        if not actions:
            continue

        for action in actions:
            res.append(action['name'])
    return res


def get_actions_by_tools(tool_names: Dict = None) -> List[str]:
    if not tool_names:
        return get_actions()

    res = []
    for tool_name, tool_info in tool_action_desc().items():
        if tool_name not in tool_names:
            continue

        actions = tool_info.get("actions")
        if not actions:
            continue

        for action in actions:
            res.append(action['name'])
    return res


def get_tool_desc():
    return tool_action_desc()


def get_tool_desc_by_name(name: str):
    return tool_action_desc().get(name, None)


def is_tool_by_name(name: str) -> bool:
    return name in ToolFactory



import sys
from abc import ABC, abstractmethod
from typing import Optional, Any, Iterator, Union, Sequence, Protocol, Iterable
from enum import Enum
from weakref import WeakSet
from dataclasses import dataclass, field
from aworld.logs.util import trace_logger


class TraceProvider(ABC):

    @abstractmethod
    def get_tracer(
            self,
            name: str,
            version: Optional[str] = None
    ) -> "Tracer":
        """Returns a `Tracer` for use by the given name.

        This function may return different `Tracer` types (e.g. a no-op tracer
        vs.  a functional tracer).

        Args:
            name: The uniquely identifiable name for instrumentation
                scope, such as instrumentation library, package, module or class name.
                ``__name__`` may not be used as this can result in
                different tracer names if the tracers are in different files.
                It is better to use a fixed string that can be imported where
                needed and used consistently as the name of the tracer.

                This should *not* be the name of the module that is
                instrumented but the name of the module doing the instrumentation.
                E.g., instead of ``"requests"``, use
                ``"opentelemetry.instrumentation.requests"``.

            version: Optional. The version string of the
                instrumenting library.  Usually this should be the same as
                ``importlib.metadata.version(instrumenting_library_name)``
        """

    @abstractmethod
    def shutdown(self) -> None:
        """Shuts down the provider and all its resources.
        This method should be called when the application is shutting down.
        """

    @abstractmethod
    def force_flush(self, timeout: Optional[float] = None) -> bool:
        """Forces all the data to be sent to the backend.
        This method should be called when the application is shutting down.
        Args:
            timeout: The maximum time to wait for the data to be sent.
        Returns:
            True if the data was sent successfully, False otherwise.
        """

    @abstractmethod
    def get_current_span(self) -> Optional["Span"]:
        """Returns the current span from the current context.
        Returns:
            The current span from the current context.
        """


class SpanType(Enum):
    """Specifies additional details on how this span relates to its parent span.
    """

    #: Default value. Indicates that the span is used internally in the
    # application.
    INTERNAL = 0

    #: Indicates that the span describes an operation that handles a remote
    # request.
    SERVER = 1

    #: Indicates that the span describes a request to some remote service.
    CLIENT = 2

    #: Indicates that the span describes a producer sending a message to a
    #: broker. Unlike client and server, there is usually no direct critical
    #: path latency relationship between producer and consumer spans.
    PRODUCER = 3

    #: Indicates that the span describes a consumer receiving a message from a
    #: broker. Unlike client and server, there is usually no direct critical
    #: path latency relationship between producer and consumer spans.
    CONSUMER = 4


AttributeValueType = Union[
    str,
    bool,
    int,
    float,
    Sequence[str],
    Sequence[bool],
    Sequence[int],
    Sequence[float],
]


class Tracer(ABC):
    """Handles span creation and in-process context propagation.
    """

    @abstractmethod
    def start_span(
            self,
            name: str,
            span_type: SpanType = SpanType.INTERNAL,
            attributes: dict[str, AttributeValueType] = None,
            start_time: Optional[int] = None,
            record_exception: bool = True,
            set_status_on_exception: bool = True,
            trace_context: Optional["TraceContext"] = None,
    ) -> "Span":
        """Starts and returns a new Span.
        Args:
            name: The name of the span.
            kind: The span's kind (relationship to parent). Note that is
                meaningful even if there is no parent.
            attributes: The span's attributes.
            start_time: Sets the start time of a span
            record_exception: Whether to record any exceptions raised within the
                context as error event on the span.
            set_status_on_exception: Only relevant if the returned span is used
                in a with/context manager. Defines whether the span status will
                be automatically set to ERROR when an uncaught exception is
                raised in the span with block. The span status won't be set by
                this mechanism if it was previously set manually.
            trace_context: The trace context to use for the span. If not
                provided, the current trace context will be used.
        """

    @abstractmethod
    def start_as_current_span(
            self,
            name: str,
            span_type: SpanType = SpanType.INTERNAL,
            attributes: dict[str, AttributeValueType] = None,
            start_time: Optional[int] = None,
            record_exception: bool = True,
            set_status_on_exception: bool = True,
            end_on_exit: bool = True,
            trace_context: Optional['TraceContext'] = None
    ) -> Iterator["Span"]:
        """Context manager for creating a new span and set it
        as the current span in this tracer's context.

        Example::

            with tracer.start_as_current_span("one") as parent:
                parent.add_event("parent's event")
                with tracer.start_as_current_span("two") as child:
                    child.add_event("child's event")
                    trace.get_current_span()  # returns child
                trace.get_current_span()      # returns parent
            trace.get_current_span()          # returns previously active span

            This can also be used as a decorator::
            @tracer.start_as_current_span("name")
            def function():

        Args:
            name: The name of the span to be created.
            kind: The span's kind (relationship to parent). Note that is
                meaningful even if there is no parent.
            attributes: The span's attributes.
            start_time: Sets the start time of a span
            record_exception: Whether to record any exceptions raised within the
                context as error event on the span.
            set_status_on_exception: Only relevant if the returned span is used
                in a with/context manager. Defines whether the span status will
                be automatically set to ERROR when an uncaught exception is
                raised in the span with block. The span status won't be set by
                this mechanism if it was previously set manually.
            end_on_exit: Whether to end the span automatically when leaving the
                context manager.
            trace_context: The trace context to use for the span. If not
                provided, the current trace context will be used.
    """


class Span(ABC):
    """A Span represents a single operation within a trace.
    """

    @abstractmethod
    def end(self, end_time: Optional[int] = None) -> None:
        """Sets the current time as the span's end time.

        The span's end time is the wall time at which the operation finished.

        Only the first call to `end` should modify the span, and
        implementations are free to ignore or raise on further calls.
        """

    @abstractmethod
    def set_attribute(self, key: str, value: Any) -> None:
        """Sets an attribute on the Span.
        Args:
            key: The attribute key.
            value: The attribute value.
        """

    @abstractmethod
    def set_attributes(self, attributes: dict[str, Any]) -> None:
        """Sets multiple attributes on the Span.
        Args:
            attributes: A dictionary of attributes to set.
        """

    @abstractmethod
    def is_recording(self) -> bool:
        """Returns whether this span will be recorded.
        Returns true if this Span is active and recording information like attributes using set_attribute.
        """

    @abstractmethod
    def record_exception(
            self,
            exception: BaseException,
            attributes: dict[str, Any] = None,
            timestamp: Optional[int] = None,
            escaped: bool = False,
    ) -> None:
        """Records an exception in the span.
        Args:
            exception: The exception to record.
            attributes: A dictionary of attributes to set on the exception event.
            timestamp: The timestamp of the exception.
            escaped: Whether the exception was escaped.
        """

    @abstractmethod
    def get_trace_id(self) -> str:
        """Returns the trace ID of the span.
        Returns:
            The trace ID of the span.
        """

    @abstractmethod
    def get_span_id(self) -> str:
        """Returns the ID of the span.
        Returns:
            The ID of the span.
        """

    def _add_to_open_spans(self) -> None:
        """Add the current span to OPEN_SPANS."""
        _OPEN_SPANS.add(self)

    def _remove_from_open_spans(self) -> None:
        """Remove the current span from OPEN_SPANS."""
        _OPEN_SPANS.discard(self)


class NoOpSpan(Span):
    """No-op implementation of `Span`."""

    def end(self, end_time: Optional[int] = None) -> None:
        pass

    def set_attribute(self, key: str, value: Any) -> None:
        pass

    def set_attributes(self, attributes: dict[str, Any]) -> None:
        pass

    def is_recording(self) -> bool:
        return False

    def record_exception(
            self,
            exception: BaseException,
            attributes: dict[str, Any] = None,
            timestamp: Optional[int] = None,
            escaped: bool = False,
    ) -> None:
        pass

    def get_trace_id(self) -> str:
        return ""

    def get_span_id(self) -> str:
        return ""


class NoOpTracer(Tracer):
    """No-op implementation of `Tracer`."""

    def start_span(
            self,
            name: str,
            span_type: SpanType = SpanType.INTERNAL,
            attributes: dict[str, AttributeValueType] = None,
            start_time: Optional[int] = None,
            record_exception: bool = True,
            set_status_on_exception: bool = True,
            trace_context: Optional["TraceContext"] = None,
    ) -> Span:
        return NoOpSpan()

    def start_as_current_span(
            self,
            name: str,
            span_type: SpanType = SpanType.INTERNAL,
            attributes: dict[str, AttributeValueType] = None,
            start_time: Optional[int] = None,
            record_exception: bool = True,
            set_status_on_exception: bool = True,
            end_on_exit: bool = True,
            trace_context: Optional['TraceContext'] = None
    ) -> Iterator[Span]:
        yield NoOpSpan()


class Carrier(Protocol):
    """Carrier is a protocol that represents a carrier for trace context.
    """

    def get(self, key: str) -> Optional[str]:
        """Returns the value of the given key from the carrier.
        Args:
            key: The key to get the value for.
        Returns:
            The value of the given key from the carrier.
        """

    def set(self, key: str, value: str) -> None:
        """Sets the value of the given key in the carrier.
        Args:
            key: The key to set the value for.
            value: The value to set.
        """

    def keys(self) -> Iterable[str]:
        """Returns an iterable of keys in the carrier.
        Returns:
            An iterable of keys in the carrier.
        """


@dataclass(frozen=True)
class TraceContext:
    """TraceContext is a class that represents a trace context.
    """
    trace_id: str
    span_id: str
    version: str = "00"
    trace_flags: str = "01"
    attributes: dict[str, Any] = field(default_factory=dict)


class Propagator(ABC):
    """Propagator is a protocol that represents a propagator for trace context.
    """

    def _get_value(self, carrier: Carrier, name: str) -> str:
        """
        Get value from carrier.
        Args:
            carrier: The carrier to get value from.
            name: The name of the value.
        Returns:
            The value of the name.
        """
        return carrier.get(name) or carrier.get('HTTP_' + name.upper().replace('-', '_'))

    @abstractmethod
    def extract(self, carrier: Carrier) -> Optional[TraceContext]:
        """Extracts a trace context from the given carrier.
        Args:
            carrier: The carrier to extract the trace context from.
        Returns:
            The trace context extracted from the carrier.
        """
    @abstractmethod
    def inject(self, trace_context: TraceContext, carrier: Carrier) -> None:
        """Injects a trace context into the given carrier.
        Args:
            trace_context: The trace context to inject.
            carrier: The carrier to inject the trace context into.
        """


_GLOBAL_TRACER_PROVIDER: Optional[TraceProvider] = None
_OPEN_SPANS: WeakSet[Span] = WeakSet()


def set_tracer_provider(provider: TraceProvider):
    """
    Set the global tracer provider.
    """
    global _GLOBAL_TRACER_PROVIDER
    _GLOBAL_TRACER_PROVIDER = provider


def get_tracer_provider() -> TraceProvider:
    """
    Get the global tracer provider.
    """
    global _GLOBAL_TRACER_PROVIDER
    if _GLOBAL_TRACER_PROVIDER is None:
        raise ValueError("No tracer provider has been set.")
    return _GLOBAL_TRACER_PROVIDER


def get_tracer_provider_silent():
    try:
        return get_tracer_provider()
    except Exception:
        return None


def log_trace_error():
    """
    Log an error with traceback information.
    """
    trace_logger.exception(
        'This is logging the trace internal error.',
        exc_info=sys.exc_info(),
    )



from aworld.logs.util import logger
from flask import Flask, render_template, jsonify
from aworld.trace.opentelemetry.memory_storage import SpanModel, TraceStorage

app = Flask(__name__, template_folder='../../web/templates')
current_storage = None
routes_setup = False



def build_trace_tree(spans: list[SpanModel]):
    spans_dict = {span.span_id: span.dict() for span in spans}
    root_spans = [span for span in spans_dict.values()
                  if span['parent_id'] is None]
    for span in spans_dict.values():
        parent_id = span['parent_id'] if span['parent_id'] else None
        if parent_id:
            parent_span = spans_dict.get(parent_id)
            if not parent_span:
                logger.warning(f"span[{parent_id}] not be exported")
                continue
            if 'children' not in parent_span:
                parent_span['children'] = []
            parent_span['children'].append(span)
    return root_spans


def setup_routes(storage: TraceStorage):
    global current_storage
    current_storage = storage

    global routes_setup

    if routes_setup:
        return app

    @app.route('/')
    def index():
        return render_template('trace_ui.html')

    @app.route('/api/traces')
    def traces():
        trace_data = []
        for trace_id in current_storage.get_all_traces():
            spans = current_storage.get_all_spans(trace_id)
            spans_sorted = sorted(spans, key=lambda x: x.start_time)
            trace_tree = build_trace_tree(spans_sorted)
            trace_data.append({
                'trace_id': trace_id,
                'root_span': trace_tree,
            })
        return jsonify(trace_data)

    @app.route('/api/traces/<trace_id>')
    def get_trace(trace_id):
        spans = current_storage.get_all_spans(trace_id)
        spans_sorted = sorted(spans, key=lambda x: x.start_time)
        trace_tree = build_trace_tree(spans_sorted)
        return jsonify({
            'trace_id': trace_id,
            'root_span': trace_tree,
        })

    routes_setup = True
    return app



# coding: utf-8
# Copyright (c) 2025 inclusionAI.

from aworld.core.tool.base import Tool, AsyncTool
from aworld.core.tool.action import ExecutableAction
from aworld.utils.common import scan_packages

scan_packages("aworld.tools", [Tool, AsyncTool, ExecutableAction])



# coding: utf-8
# Copyright (c) 2025 inclusionAI.
from typing import Dict, List

from aworld.core.agent.agent_desc import agent_handoffs_desc
from aworld.core.agent.base import AgentFactory
from aworld.core.agent.llm_agent import Agent
from aworld.core.common import ActionModel, Observation
from aworld.core.context.base import Context
from aworld.logs.util import logger

SEQUENCE = "sequence"
SOCIAL = "social"
SEQUENCE_EVENT = "sequence_event"
SOCIAL_EVENT = "social_event"


class Swarm(object):
    """Multi-agent topology.

    Examples:
        >>> agent1 = Agent(name='agent1'); agent2 = Agent(name='agent2'); agent3 = Agent(name='agent3')
        # sequencial
        >>> Swarm(agent1, agent2, agent3)
        # social
        >>> Swarm((agent1, agent2), (agent1, agent3), (agent2, agent3), sequence=False)
    """

    def __init__(self, *args, root_agent: Agent = None, sequence: bool = True, max_steps: int = 0, **kwargs):
        self.communicate_agent = root_agent
        if root_agent and root_agent not in args:
            self._topology = [root_agent] + list(args)
        else:
            self._topology = args
        self._ext_params = kwargs
        self.sequence = sequence
        self.max_steps = max_steps
        self.initialized = False
        self._finished = False
        self._cur_step = 0
        self._event_driven = kwargs.get('event_driven', False)
        for agent in self._topology:
            if isinstance(agent, Agent):
                agent = [agent]
            for a in agent:
                if a and a.event_driven:
                    self._event_driven = True
                    break
            if self._event_driven:
                break

    @property
    def event_driven(self):
        return self._event_driven

    @event_driven.setter
    def event_driven(self, event_driven):
        self._event_driven = event_driven

    def _init(self, **kwargs):
        """Swarm init, build the agent or agent pairs to the topology of agents."""
        # prebuild
        valid_agent_pair = []
        for pair in self._topology:
            if isinstance(pair, (list, tuple)):
                self.topology_type = SOCIAL
                # (agent1, agent2)
                if len(pair) != 2:
                    raise RuntimeError(f"{pair} is not a pair value, please check it.")
                elif not isinstance(pair[0], Agent) or not isinstance(pair[1], Agent):
                    raise RuntimeError(f"agent in {pair} is not a base agent instance, please check it.")
                valid_agent_pair.append(pair)
            else:
                # agent
                if not isinstance(pair, Agent):
                    raise RuntimeError(f"agent in {pair} is not a base agent instance, please check it.")
                self.topology_type = SEQUENCE
                valid_agent_pair.append((pair,))

        if not valid_agent_pair:
            logger.warning("no valid agent pair to build graph.")
            return

        # Agent that communicate with the outside world, the default is the first if the root agent is None.
        if self.communicate_agent is None:
            self.communicate_agent: Agent = valid_agent_pair[0][0]
        # agents in swarm.
        self.agents: Dict[str, Agent] = dict()
        self.ordered_agents: List[Agent] = []

        # agent handoffs build.
        for pair in valid_agent_pair:
            if self.sequence or self.topology_type == SEQUENCE:
                self.ordered_agents.append(pair[0])
                if len(pair) == 2:
                    self.ordered_agents.append(pair[1])

            if pair[0] not in self.agents:
                self.agents[pair[0].name()] = pair[0]
            if pair[0].name() not in AgentFactory:
                AgentFactory._cls[pair[0].name()] = pair[0].__class__
                AgentFactory._desc[pair[0].name()] = pair[0].desc()
                AgentFactory._agent_conf[pair[0].name()] = pair[0].conf
                AgentFactory._agent_instance[pair[0].name()] = pair[0]
            elif pair[0].desc():
                AgentFactory._desc[pair[0].name()] = pair[0].desc()

            if len(pair) == 1:
                continue

            if pair[1] not in self.agents:
                self.agents[pair[1].name()] = pair[1]
                if pair[1].name() not in AgentFactory:
                    AgentFactory._cls[pair[1].name()] = pair[1].__class__
                    AgentFactory._desc[pair[1].name()] = pair[1].desc()
                    AgentFactory._agent_conf[pair[1].name()] = pair[1].conf
                    AgentFactory._agent_instance[pair[1].name()] = pair[1]
                elif pair[1].desc():
                    AgentFactory._desc[pair[1].name()] = pair[1].desc()

            if self.topology_type == SOCIAL:
                # need to explicitly set handoffs in the agent
                pair[0].handoffs.append(pair[1].name())

        if self.sequence:
            self.topology_type = SEQUENCE

        self._cur_step = 1
        # event driven
        if self.event_driven:
            for _, agent in self.agents.items():
                agent.event_driven = True
            if self.topology_type == SEQUENCE:
                self.topology_type = SEQUENCE_EVENT
            elif self.topology_type == SOCIAL:
                self.topology_type = SOCIAL_EVENT

    def reset(self, content: str, context: Context = None, tools: List[str] = []):
        """Resets the initial internal state, and init supported tools in agent in swarm.

        Args:
            tools: Tool names that all agents in the swarm can use.
        """
        # can use the tools in the agents in the swarm as a global
        self.tools = tools

        self._init()
        if not self.agents:
            logger.warning("No valid agent in swarm.")
            return

        self.cur_agent = self.communicate_agent

        if context is None:
            context = Context.instance()

        for agent in self.agents.values():
            if agent.need_reset:
                agent.context = context
                agent.reset({"task": content,
                             "tool_names": agent.tool_names,
                             "agent_names": agent.handoffs,
                             "mcp_servers": agent.mcp_servers})
            # global tools
            agent.tool_names.extend(self.tools)
        self.initialized = True

    def _check(self):
        if not self.initialized:
            self.reset('')

    def handoffs_desc(self, agent_name: str = None, use_all: bool = False):
        """Get agent description by name for handoffs.

        Args:
            agent_name: Agent unique name.
        Returns:
            Description of agent dict.
        """
        self._check()
        agent: Agent = self.agents.get(agent_name, None)
        return agent_handoffs_desc(agent, use_all)

    def action_to_observation(self, policy: List[ActionModel], observation: List[Observation], strategy: str = None):
        """Based on the strategy, transform the agent's policy into an observation, the case of the agent as a tool.

        Args:
            policy: Agent policy based some messages.
            observation: History of the current observable state in the environment.
            strategy: Transform strategy, default is None. enum?
        """
        self._check()

        if not policy:
            logger.warning("no agent policy, will return origin observation.")
            # get the latest one
            if not observation:
                raise RuntimeError("no observation and policy to transform in swarm, please check your params.")
            return observation[-1]

        if not strategy:
            # default use the first policy
            policy_info = policy[0].policy_info

            if not observation:
                res = Observation(content=policy_info)
            else:
                res = observation[-1]
                if res.content is None:
                    res.content = ''

                if policy_info:
                    res.content += policy_info
            return res
        else:
            logger.warning(f"{strategy} not supported now.")

    def supported_tools(self):
        """Tool names that can be used by all agents in Swarm."""
        self._check()
        return self.tools

    @property
    def cur_step(self) -> int:
        return self._cur_step

    @cur_step.setter
    def cur_step(self, step):
        self._cur_step = step

    @property
    def finished(self) -> bool:
        """Need all agents in a finished state."""
        self._check()
        if not self._finished:
            self._finished = all([agent.finished for _, agent in self.agents.items()])
        return self._finished

    @finished.setter
    def finished(self, finished):
        self._finished = finished




# auto generated
class VersionInfo(object):
    @property
    def build_date(self):
      return "2025-06-10 08:16:54"

    @property
    def version(self):
      return "0.2.4"

    @property
    def build_user(self):
      return "ganrunsheng"



import streamlit as st
from dotenv import load_dotenv
import logging
import os

load_dotenv(os.path.join(os.getcwd(), ".env"))

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

chat = st.Page("chat.py", title="Chat", icon=":material/message:")
trace = st.Page("trace.py", title="Trace")

pg = st.navigation([chat, trace], position="hidden")
pg.run()



import asyncio
import types
import inspect
import contextvars
from typing import Union, Optional, Any, Type, Sequence, Callable, Iterable
from aworld.trace.base import (
    AttributeValueType,
    Span, Tracer,
    NoOpTracer,
    get_tracer_provider,
    get_tracer_provider_silent,
    log_trace_error
)
from aworld.trace.span_cosumer import SpanConsumer
from aworld.version_gen import __version__
from aworld.trace.auto_trace import AutoTraceModule, install_auto_tracing
from aworld.trace.stack_info import get_user_stack_info
from aworld.trace.constants import (
    ATTRIBUTES_MESSAGE_KEY,
    ATTRIBUTES_MESSAGE_RUN_TYPE_KEY,
    ATTRIBUTES_MESSAGE_TEMPLATE_KEY,
    RunType
)
from aworld.trace.msg_format import (
    chunks_formatter,
    warn_formatting,
    FStringAwaitError,
    KnownFormattingError,
    warn_fstring_await
)
from aworld.trace.function_trace import trace_func
from .opentelemetry.opentelemetry_adapter import configure_otlp_provider
from aworld.logs.util import logger


def trace_configure(provider: str = "otlp",
                    backends: Sequence[str] = None,
                    base_url: str = None,
                    write_token: str = None,
                    span_consumers: Optional[Sequence[SpanConsumer]] = None,
                    **kwargs
                    ) -> None:
    """
    Configure the trace provider.
    Args:
        provider: The trace provider to use.
        backends: The trace backends to use.
        base_url: The base URL of the trace backend.
        write_token: The write token of the trace backend.
        span_consumers: The span consumers to use.
        **kwargs: Additional arguments to pass to the trace provider.
    Returns:
        None
    """
    exist_provider = get_tracer_provider_silent()
    if exist_provider:
        logger.info("Trace provider already configured, shutting down...")
        exist_provider.shutdown()
    if provider == "otlp":
        configure_otlp_provider(
            backends=backends, base_url=base_url, write_token=write_token, span_consumers=span_consumers, **kwargs)
    else:
        raise ValueError(f"Unknown trace provider: {provider}")


class TraceManager:
    """
    TraceManager is a class that provides a way to trace the execution of a function.
    """

    def __init__(self, tracer_name: str = None) -> None:
        self._tracer_name = tracer_name or "aworld"
        self._version = __version__

    def _create_auto_span(self,
                          name: str,
                          attributes: dict[str, AttributeValueType] = None
                          ) -> Span:
        """
        Create a auto trace span with the given name and attributes.
        """
        try:
            tracer = get_tracer_provider().get_tracer(
                name=self._tracer_name, version=self._version)
            return ContextSpan(span_name=name, tracer=tracer, attributes=attributes)
        except Exception:
            return ContextSpan(span_name=name, tracer=NoOpTracer(), attributes=attributes)

    def get_current_span(self) -> Span:
        """
        Get the current span.
        """
        try:
            return get_tracer_provider().get_current_span()
        except Exception:
            return None

    def new_manager(self, tracer_name_suffix: str = None) -> "TraceManager":
        """
        Create a new TraceManager with the given tracer name suffix.
        """
        tracer_name = self._tracer_name if not tracer_name_suffix else f"{self._tracer_name}.{tracer_name_suffix}"
        return TraceManager(tracer_name=tracer_name)

    def auto_tracing(self,
                     modules: Union[Sequence[str], Callable[[AutoTraceModule], bool]],
                     min_duration: float) -> None:
        """
        Automatically trace the execution of a function.
        Args:
            modules: A list of module names or a callable that takes a `AutoTraceModule` and returns a boolean.
            min_duration: The minimum duration of a function to be traced.
        Returns:
            None
        """
        install_auto_tracing(self, modules, min_duration)

    def span(self,
             msg_template: str,
             attributes: dict[str, AttributeValueType] = None,
             *,
             span_name: str = None,
             run_type: RunType = RunType.OTHER) -> "ContextSpan":

        try:
            attributes = attributes or {}
            stack_info = get_user_stack_info()
            merged_attributes = {**stack_info, **attributes}
            # Retrieve stack information of user code and add it to the attributes

            if any(c in msg_template for c in ('{', '}')):
                fstring_frame = inspect.currentframe().f_back
            else:
                fstring_frame = None
            log_message, extra_attrs, msg_template = format_span_msg(
                msg_template,
                merged_attributes,
                fstring_frame=fstring_frame,
            )
            merged_attributes[ATTRIBUTES_MESSAGE_KEY] = log_message
            merged_attributes.update(extra_attrs)
            merged_attributes[ATTRIBUTES_MESSAGE_TEMPLATE_KEY] = msg_template
            merged_attributes[ATTRIBUTES_MESSAGE_RUN_TYPE_KEY] = run_type.value
            span_name = span_name or msg_template
            tracer = get_tracer_provider().get_tracer(
                name=self._tracer_name, version=self._version)
            return ContextSpan(span_name=span_name,
                               tracer=tracer,
                               attributes=merged_attributes)
        except Exception:
            log_trace_error()
            return ContextSpan(span_name=span_name, tracer=NoOpTracer(), attributes=attributes)

    def func_span(self,
                  msg_template: Union[str, Callable] = None,
                  *,
                  attributes: dict[str, AttributeValueType] = None,
                  span_name: str = None,
                  extract_args: Union[bool, Iterable[str]] = False,
                  **kwargs) -> Callable:
        """
        A decorator that traces the execution of a function.
        Args:
            msg_template: The message template to use.
            attributes: The attributes to add to the span.
            span_name: The name of the span.
            extract_args: Whether to extract arguments from the function call.
            **kwargs: Additional attributes to add to the span.
        Returns:
            A decorator that traces the execution of a function.
        """
        if callable(msg_template):
            #  @trace_func
            #  def foo():
            return self.func_span()(msg_template)

        attributes = attributes or {}
        attributes.update(kwargs)
        return trace_func(self, msg_template, attributes, span_name, extract_args)


class ContextSpan(Span):
    """A context manager that wraps an existing `Span` object.
    This class provides a way to use a `Span` object as a context manager.
    When the context manager is entered, it returns the `Span` itself.
    When the context manager is exited, it calls `end` on the `Span`.
    Args:
        span: The `Span` object to wrap.
    """

    def __init__(self,
                 span_name: str,
                 tracer: Tracer,
                 attributes: dict[str, AttributeValueType] = None) -> None:
        self._span_name = span_name
        self._tracer = tracer
        self._attributes = attributes
        self._span: Span = None
        self._coro_context = None

    def _start(self):
        if self._span is not None:
            return

        self._span = self._tracer.start_span(
            name=self._span_name,
            attributes=self._attributes,
        )

    def __enter__(self) -> "Span":
        self._start()
        return self

    def __exit__(
            self,
            exc_type: Optional[Type[BaseException]],
            exc_val: Optional[BaseException],
            traceback: Optional[Any],
    ) -> None:
        """Ends context manager and calls `end` on the `Span`."""
        self._handle_exit(exc_type, exc_val, traceback)

    async def __aenter__(self) -> "Span":
        self._coro_context = contextvars.copy_context()
        await asyncio.get_event_loop().run_in_executor(None, lambda: self._coro_context.run(self._start)
                                                       )
        return self

    async def __aexit__(
            self,
            exc_type: Optional[Type[BaseException]],
            exc_val: Optional[BaseException],
            traceback: Optional[Any],
    ) -> None:
        await asyncio.get_event_loop().run_in_executor(None,
                                                       lambda: self._coro_context.run(lambda: self._handle_exit(exc_type, exc_val, traceback)))

    def _handle_exit(
            self,
            exc_type: Optional[Type[BaseException]],
            exc_val: Optional[BaseException],
            traceback: Optional[Any],
    ) -> None:
        try:
            if self._span and self._span.is_recording() and isinstance(exc_val, BaseException):
                self._span.record_exception(exc_val, escaped=True)
        except ValueError as e:
            logger.warning(f"Failed to record_exception: {e}")
        finally:
            if self._span:
                self._span.end()

    def end(self, end_time: Optional[int] = None) -> None:
        if self._span:
            self._span.end(end_time)

    def set_attribute(self, key: str, value: AttributeValueType) -> None:
        if self._span:
            self._span.set_attribute(key, value)

    def set_attributes(self, attributes: dict[str, AttributeValueType]) -> None:
        if self._span:
            self._span.set_attributes(attributes)

    def is_recording(self) -> bool:
        if self._span:
            return self._span.is_recording()
        return False

    def record_exception(
            self,
            exception: BaseException,
            attributes: dict[str, Any] = None,
            timestamp: Optional[int] = None,
            escaped: bool = False,
    ) -> None:
        if self._span:
            self._span.record_exception(
                exception, attributes, timestamp, escaped)

    def get_trace_id(self) -> str:
        if self._span:
            return self._span.get_trace_id()

    def get_span_id(self) -> str:
        if self._span:
            return self._span.get_span_id()


def format_span_msg(
        format_string: str,
        kwargs: dict[str, Any],
        fstring_frame: types.FrameType = None,
) -> tuple[str, dict[str, Any], str]:
    """ Returns
    1. The formatted message.
    2. A dictionary of extra attributes to add to the span/log.
         These can come from evaluating values in f-strings.
    3. The final message template, which may differ from `format_string` if it was an f-string.
    """
    try:
        chunks, extra_attrs, new_template = chunks_formatter.chunks(
            format_string,
            kwargs,
            fstring_frame=fstring_frame
        )
        return ''.join(chunk['v'] for chunk in chunks), extra_attrs, new_template
    except KnownFormattingError as e:
        warn_formatting(str(e) or str(e.__cause__))
    except FStringAwaitError as e:
        warn_fstring_await(str(e))
    except Exception:
        log_trace_error()

    # Formatting failed, so just use the original format string as the message.
    return format_string, {}, format_string



# coding: utf-8
# Copyright (c) 2025 inclusionAI.

import os
import sys
import re
import subprocess

from aworld.version_gen import generate_version_info, __version__

from setuptools import setup, find_packages
from setuptools.command.sdist import sdist
from setuptools.command.install import install
from setuptools.dist import Distribution

from aworld.logs.util import logger

version_template = """
# auto generated
class VersionInfo(object):
    @property
    def build_date(self):
      return "{BUILD_DATE}"

    @property
    def version(self):
      return "{BUILD_VERSION}"

    @property
    def build_user(self):
      return "ganrunsheng"
"""


def check_output(cmd):
    import subprocess
    output = subprocess.check_output(cmd)
    return output.decode("utf-8")


def get_build_date():
    import datetime
    import time
    ts = time.time()
    return datetime.datetime.fromtimestamp(ts).strftime('%Y-%m-%d %H:%M:%S')


def build_version_template():
    import getpass
    return version_template.format(BUILD_USER=getpass.getuser(),
                                   BUILD_VERSION=__version__,
                                   BUILD_DATE=get_build_date())


def call_process(cmd, raise_on_error=True, logging=True):
    if isinstance(cmd, str):
        shell = True
    else:
        shell = False  # cmd should be list of args

    try:
        subprocess.check_call(cmd, shell=shell, timeout=60)
    except subprocess.CalledProcessError as e:
        if raise_on_error:
            raise e
        logger.error(f"Fail to execute {cmd}, {e}")
        return e.returncode

    if logging:
        logger.info(f"Successfully execute: {cmd}")
    return 0


class AWorldPackage(sdist):
    def run(self):
        import os
        home = os.path.join(os.path.dirname(__file__), "aworld")
        with open(os.path.join(home, "version.py"), "w") as f:
            version_info = build_version_template()
            f.write(version_info)

        generate_version_info(scenario="AWORLD_SDIST")
        sdist.run(self)


class AWorldInstaller(install):
    EXTRA_ENV = "AWORLD_EXTRA"
    BASE = "framework"
    BASE_OPT = "optional"

    def __init__(self, *args, **kwargs):
        super(AWorldInstaller, self).__init__(*args, **kwargs)
        self._requirements = parse_requirements("aworld/requirements.txt")
        self._extra = os.getenv(self.EXTRA_ENV)
        logger.info(f"{os.getcwd()}: Install AWORLD using extra: {self._extra}")

    def run(self):
        # 1. build wheel using this setup.py, thus using the right install_requires according to ALPS_EXTRA
        # 2. install this wheel into pip
        install.run(self)

        reqs = self._requirements.get(self.BASE, [])
        self._install_reqs(reqs, ignore_error=True)

        # install optional requirements here since pip install doesn't ignore requirement error
        reqs = self._requirements.get(self.BASE_OPT, [])
        self._install_reqs(reqs, ignore_error=True)

    def _contains_module(self, module):
        if self._extra is None:
            return False

        modules = [mod.strip() for mod in self._extra.split(',')]
        try:
            modules.index(module)
            return True
        except ValueError:
            return False

    @staticmethod
    def _install_reqs(reqs, ignore_error=False, no_deps=False):
        info = "--no-deps" if no_deps else ""
        if ignore_error:
            # install requirements one by one
            for req in reqs:
                try:
                    cmd = f"{sys.executable} -m pip install {info} {req}"
                    call_process(cmd)
                    logger.info(f"Installing optional package {req} have succeeded.")
                except:
                    logger.warning(f"Installing optional package {req} is failed, Ignored.")  # ignore
        elif reqs:
            cmd = f"{sys.executable} -m pip install {info} {' '.join(reqs)}"
            call_process(cmd)
            logger.info(f"Packages {str(reqs)} have been installed.")


def parse_requirements(req_fname):
    requirements = {}
    module_name = "unknown"
    for line in open(req_fname, "r"):
        match = re.match(r'#+\s+\[(\w+)\]\s+#+', line.strip())
        if match:
            # the beginning of a module
            module_name = match.group(1)
        else:
            req = line.strip()
            if not req or req.startswith("#"):
                continue

            # it's a requirement, strip trailing comments
            pos = req.find("#")
            if pos > 0:
                req = req[:pos]
                req = req.strip()

            if module_name not in requirements:
                requirements[module_name] = []
            requirements[module_name].append(req)

    return requirements


def get_install_requires(extra, requirements):
    modules = [AWorldInstaller.BASE]
    if extra is None:
        # old style of `pip install alps`, install all requirements for compatibility
        for mod in requirements:
            if mod in [AWorldInstaller.BASE, AWorldInstaller.BASE_OPT]:
                continue
            modules.append(mod)
    else:
        for mod in extra.split(','):
            mod = mod.strip()
            if mod != AWorldInstaller.BASE:
                modules.append(mod)

    install_reqs = []
    for mod in modules:
        install_reqs.extend(requirements.get(mod, []))
    return install_reqs


def get_python_requires():
    return ">=3.11"


class BinaryDistribution(Distribution):
    """This class is needed in order to create OS specific wheels."""

    @staticmethod
    def has_ext_modules():
        return True


requirements = parse_requirements('aworld/requirements.txt')
extra = os.getenv(AWorldInstaller.EXTRA_ENV, None)

setup(
    name='aworld',
    version=__version__,
    description='Ant Agent Package',
    url='https://github.com/inclusionAI/AWorld',
    author='Ant AI',
    author_email='',
    long_description='',
    long_description_content_type="text/markdown",
    packages=find_packages(where=".", exclude=['tests', 'tests.*', '*.tests', '*.tests.*', '*.test', '*.test.*']),
    package_data={
        'aworld': [
            'virtual_environments/browsers/script/*.js',
            'dataset/gaia/gaia.npy',
            'requirements.txt',
            'config/*.yaml',
            'config/*.json',
            'web/templates/*.html'
        ]
    },
    license='MIT',
    platforms=["any"],
    keywords=['multi-agent', 'agent', 'environment', "tool", 'sandbox'],
    cmdclass={
        'sdist': AWorldPackage,
        'install': AWorldInstaller,
    },
    install_requires=get_install_requires(extra, requirements),
    python_requires=get_python_requires(),
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',

        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
    ],
    entry_points={
        'console_scripts': [
            'aworld = aworld.__main__:main',
        ]
    })



from uuid import uuid4
from contextvars import ContextVar
from types import MappingProxyType

_BAGGAGE_KEY = "aworld.baggage." + str(uuid4())

_BAGGAGE_CONTEXT = ContextVar(_BAGGAGE_KEY, default=None)


class BaggageContext:
    """
    Baggage context.
    """

    @staticmethod
    def get_baggage() -> dict:
        """
        Get the baggage. This is a read-only view of the baggage.
        Returns:
            The baggage.
        """
        baggage = _BAGGAGE_CONTEXT.get()
        if isinstance(baggage, dict):
            return MappingProxyType(baggage)
        return {}

    @staticmethod
    def get_baggage_value(key: str):
        """
        Get the value for a key from baggage.
        Args:
            key: The key of the value to retrieve.
        Returns:
            The baggage value.
        """
        baggage = BaggageContext.get_baggage()
        if key:
            return baggage.get(key)
        return None

    @staticmethod
    def set_baggage(key: str, value: object):
        """
        Set the value for a key in baggage.
        Args:
            key: The key of the value to set.
            value: The value to set.
        """
        baggage = BaggageContext.get_baggage().copy()
        baggage[key] = value
        _BAGGAGE_CONTEXT.set(baggage)


import os
import sys
import traceback

from dotenv import load_dotenv
from mcp.server.fastmcp import FastMCP
from pydantic import Field

from aworld.config.conf import AgentConfig
from aworld.logs.util import logger
from aworld.models.llm import call_llm_model, get_llm_model

# Initialize MCP server
mcp = FastMCP("reasoning-server")


@mcp.tool(
    description="Perform complex problem reasoning using powerful reasoning model."
)
def complex_problem_reasoning(
    question: str = Field(
        description="The input question for complex problem reasoning,"
        + " such as math and code contest problem",
    ),
    original_task: str = Field(
        default="",
        description="The original task description."
        + " This argument could be fetched from the <task>TASK</task> tag",
    ),
) -> str:
    """
    Perform complex problem reasoning using Powerful Reasoning model,
    such as riddle, game or competition-level STEM(including code) problems.

    Args:
        question: The input question for complex problem reasoning
        original_task: The original task description (optional)

    Returns:
        str: The reasoning result from the model
    """
    try:
        # Prepare the prompt with both the question and original task if provided
        prompt = question
        if original_task:
            prompt = f"Original Task: {original_task}\n\nQuestion: {question}"

        # Call the LLM model for reasoning
        response = call_llm_model(
            llm_model=get_llm_model(
                conf=AgentConfig(
                    llm_provider="openai",
                    llm_model_name="anthropic/claude-3.7-sonnet:thinking",
                    llm_api_key=os.getenv("LLM_API_KEY", "your_openai_api_key"),
                    llm_base_url=os.getenv("LLM_BASE_URL", "your_openai_base_url"),
                )
            ),
            messages=[
                {
                    "role": "system",
                    "content": (
                        "You are an expert at solving complex problems including math,"
                        " code contests, riddles, and puzzles."
                        " Provide detailed step-by-step reasoning and a clear final answer."
                    ),
                },
                {"role": "user", "content": prompt},
            ],
            temperature=float(os.getenv("LLM_TEMPERATURE", "0.3")),
        )

        # Extract the reasoning result
        reasoning_result = response.content

        logger.info("Complex reasoning completed successfully")
        return reasoning_result

    except Exception as e:
        logger.error(f"Error in complex problem reasoning: {traceback.format_exc()}")
        return f"Error performing reasoning: {str(e)}"


def main():
    load_dotenv()
    print("Starting Reasoning MCP Server...", file=sys.stderr)
    mcp.run(transport="stdio")


# Make the module callable
def __call__():
    """
    Make the module callable for uvx.
    This function is called when the module is executed directly.
    """
    main()


sys.modules[__name__].__call__ = __call__

# Run the server when the script is executed directly
if __name__ == "__main__":
    main()



# coding: utf-8
# Copyright (c) 2025 inclusionAI.

import abc
import uuid


import aworld.trace as trace

from typing import Generic, TypeVar, Dict, Any, List, Tuple, Union

from pydantic import BaseModel

from aworld.config.conf import AgentConfig, load_config, ConfigDict
from aworld.core.common import Observation, ActionModel
from aworld.core.context.base import Context
from aworld.core.event.base import Message
from aworld.core.factory import Factory
from aworld.logs.util import logger
from aworld.sandbox.base import Sandbox

from aworld.utils.common import convert_to_snake, replace_env_variables

INPUT = TypeVar('INPUT')
OUTPUT = TypeVar('OUTPUT')


def is_agent_by_name(name: str) -> bool:
    return name in AgentFactory


def is_agent(policy: ActionModel) -> bool:
    return is_agent_by_name(policy.tool_name) or (not policy.tool_name and not policy.action_name)


class AgentStatus:
    # Init status
    START = 0
    # Agent is running for monitor or collection
    RUNNING = 1
    # Agent reject the task
    REJECT = 2
    # Agent is idle
    IDLE = 3
    # Agent meets exception
    ERROR = 4
    # End of one agent step
    DONE = 5
    # End of one task step
    FINISHED = 6


class AgentResult(BaseModel):
    current_state: Any
    actions: List[ActionModel]
    is_call_tool: bool = True


class MemoryModel(BaseModel):
    # TODO: memory module
    message: Dict = {}
    tool_calls: Any = None
    content: Any = None


class BaseAgent(Generic[INPUT, OUTPUT]):
    __metaclass__ = abc.ABCMeta

    def __init__(self,
                 conf: Union[Dict[str, Any],
                             ConfigDict, AgentConfig],
                 sandbox: Sandbox = None,
                 mcp_servers: List[str] = [],
                 mcp_config: Dict[str, Any] = {},
                 **kwargs
                 ):
        self.conf = conf
        if isinstance(conf, ConfigDict):
            pass
        elif isinstance(conf, Dict):
            self.conf = ConfigDict(conf)
        elif isinstance(conf, AgentConfig):
            # To add flexibility
            self.conf = ConfigDict(conf.model_dump())
        else:
            logger.warning(f"Unknown conf type: {type(conf)}")

        self._name = kwargs.pop("name", self.conf.get(
            "name", convert_to_snake(self.__class__.__name__)))
        self._desc = kwargs.pop("desc") if kwargs.get(
            "desc") else self.conf.get('desc', '')
        # Unique flag based agent name
        self.id = f"{self.name()}_{uuid.uuid1().hex[0:6]}"
        self.task = None
        # An agent can use the tool list
        self.tool_names: List[str] = kwargs.pop("tool_names", [])
        # An agent can delegate tasks to other agent
        self.handoffs: List[str] = kwargs.pop("agent_names", [])
        # Supported MCP server
        self.mcp_servers: List[str] = mcp_servers
        self.mcp_config: Dict[str, Any] = replace_env_variables(mcp_config)
        self.trajectory: List[Tuple[INPUT, Dict[str, Any], AgentResult]] = []
        # all tools that the agent can use. note: string name/id only
        self.tools = []
        self.context = Context.instance()
        self.state = AgentStatus.START
        self._finished = True
        # todo sandbox
        self.sandbox = sandbox or Sandbox(
            mcp_servers=self.mcp_servers, mcp_config=self.mcp_config)

        for k, v in kwargs.items():
            setattr(self, k, v)

    def name(self) -> str:
        return self._name

    def desc(self) -> str:
        return self._desc

    def run(self, observation: Observation, info: Dict[str, Any] = {}, **kwargs) -> Message:
        with trace.span(self._name, run_type=trace.RunType.AGNET) as agent_span:
            self.pre_run()
            result = self.policy(observation, info, **kwargs)
            final_result = self.post_run(result, observation)
            return final_result if final_result else result

    async def async_run(self, observation: Observation, info: Dict[str, Any] = {}, **kwargs) -> Message:
        with trace.span(self._name, run_type=trace.RunType.AGNET) as agent_span:
            await self.async_pre_run()
            result = await self.async_policy(observation, info, **kwargs)
            final_result = await self.async_post_run(result, observation)
            return final_result if final_result else result

    @abc.abstractmethod
    def policy(self, observation: INPUT, info: Dict[str, Any] = None, **kwargs) -> OUTPUT:
        """The strategy of an agent can be to decide which tools to use in the environment, or to delegate tasks to other agents.

        Args:
            observation: The state observed from tools in the environment.
            info: Extended information is used to assist the agent to decide a policy.
        """

    @abc.abstractmethod
    async def async_policy(self, observation: INPUT, info: Dict[str, Any] = None, **kwargs) -> OUTPUT:
        """The strategy of an agent can be to decide which tools to use in the environment, or to delegate tasks to other agents.

        Args:
            observation: The state observed from tools in the environment.
            info: Extended information is used to assist the agent to decide a policy.
        """

    def reset(self, options: Dict[str, Any]):
        """Clean agent instance state and reset."""
        if options is None:
            options = {}
        self.task = options.get("task")
        self.tool_names = options.get("tool_names", [])
        self.handoffs = options.get("agent_names", [])
        self.mcp_servers = options.get("mcp_servers", [])
        self.tools = []
        self.trajectory = []
        self._finished = True

    async def async_reset(self, options: Dict[str, Any]):
        """Clean agent instance state and reset."""
        self.task = options.get("task")

    @property
    def finished(self) -> bool:
        """Agent finished the thing, default is True."""
        return self._finished

    def pre_run(self):
        pass

    def post_run(self, policy_result: OUTPUT, input: INPUT) -> Message:
        pass

    async def async_pre_run(self):
        pass

    async def async_post_run(self, policy_result: OUTPUT, input: INPUT) -> Message:
        pass


class AgentManager(Factory):
    def __init__(self, type_name: str = None):
        super(AgentManager, self).__init__(type_name)
        self._agent_conf = {}
        self._agent_instance = {}

    def __call__(self, name: str = None, *args, **kwargs):
        if name is None:
            return self

        conf = self._agent_conf.get(name)
        if not conf:
            logger.warning(f"{name} not find conf in agent factory")
            conf = dict()
        elif isinstance(conf, BaseModel):
            conf = conf.model_dump()

        user_conf = kwargs.pop('conf', None)
        if user_conf:
            if isinstance(user_conf, BaseModel):
                conf.update(user_conf.model_dump())
            elif isinstance(user_conf, dict):
                conf.update(user_conf)
            else:
                logger.warning(
                    f"Unknown conf type: {type(user_conf)}, ignored!")

        conf['name'] = name
        conf = ConfigDict(conf)
        if name in self._cls:
            agent = self._cls[name](conf=conf, **kwargs)
            self._agent_instance[name] = agent
        else:
            raise ValueError(f"Can not find {name} agent!")
        return agent

    def desc(self, name: str) -> str:
        if self._agent_instance.get(name, None) and self._agent_instance[name].desc:
            return self._agent_instance[name].desc
        return self._desc.get(name, "")

    def agent_instance(self, name: str) -> BaseAgent | None:
        if self._agent_instance.get(name, None):
            return self._agent_instance[name]
        return None

    def register(self, name: str, desc: str, conf_file_name: str = None, **kwargs):
        """Register a tool to tool factory.

        Args:
            name: Tool name
            desc: Tool description
            supported_action: Tool abilities
            conf_file_name: Default tool config
        """
        res = super(AgentManager, self).register(name, desc, **kwargs)
        conf_file_name = conf_file_name if conf_file_name else f"{name}.yaml"
        conf = load_config(conf_file_name, kwargs.get("dir"))
        if not conf:
            logger.warning(f"{conf_file_name} not find, will use default")
            # use general tool config
            conf = AgentConfig().model_dump()
        self._agent_conf[name] = conf
        return res


AgentFactory = AgentManager("agent_type")



# coding: utf-8
# Copyright (c) 2025 inclusionAI.
import os
try:
    import aworld.tools
except:
    pass
from aworld.utils import import_package
import_package("dotenv", install_name="python-dotenv")

from dotenv import load_dotenv

sucess = load_dotenv()
if not sucess:
    load_dotenv(os.path.join(os.getcwd(), '.env'))
